# coding=utf-8

# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for :class:`ironic.conductor.task_manager`."""

import futurist
import mock
from oslo_utils import uuidutils

from ironic.common import driver_factory
from ironic.common import exception
from ironic.common import fsm
from ironic.common import states
from ironic.conductor import notification_utils
from ironic.conductor import task_manager
from ironic import objects
from ironic.objects import fields
from ironic.tests import base as tests_base
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


@mock.patch.object(objects.Node, 'get')
@mock.patch.object(objects.Node, 'release')
@mock.patch.object(objects.Node, 'reserve')
@mock.patch.object(driver_factory, 'build_driver_for_task')
@mock.patch.object(objects.Port, 'list_by_node_id')
@mock.patch.object(objects.Portgroup, 'list_by_node_id')
@mock.patch.object(objects.VolumeConnector, 'list_by_node_id')
@mock.patch.object(objects.VolumeTarget, 'list_by_node_id')
class TaskManagerTestCase(db_base.DbTestCase):
    def setUp(self):
        super(TaskManagerTestCase, self).setUp()
        self.host = 'test-host'
        self.config(host=self.host)
        self.config(node_locked_retry_attempts=1, group='conductor')
        self.config(node_locked_retry_interval=0, group='conductor')
        self.node = obj_utils.create_test_node(self.context)
        self.future_mock = mock.Mock(spec=['cancel', 'add_done_callback'])

    def test_excl_lock(self, get_voltgt_mock, get_volconn_mock,
                       get_portgroups_mock, get_ports_mock,
                       build_driver_mock, reserve_mock, release_mock,
                       node_get_mock):
        reserve_mock.return_value = self.node
        with task_manager.TaskManager(self.context, 'fake-node-id') as task:
            self.assertEqual(self.context, task.context)
            self.assertEqual(self.node, task.node)
            self.assertEqual(get_ports_mock.return_value, task.ports)
            self.assertEqual(get_portgroups_mock.return_value, task.portgroups)
            self.assertEqual(get_volconn_mock.return_value,
                             task.volume_connectors)
            self.assertEqual(get_voltgt_mock.return_value, task.volume_targets)
            self.assertEqual(build_driver_mock.return_value, task.driver)
            self.assertFalse(task.shared)
            build_driver_mock.assert_called_once_with(task)

        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        reserve_mock.assert_called_once_with(self.context, self.host,
                                             'fake-node-id')
        get_ports_mock.assert_called_once_with(self.context, self.node.id)
        get_portgroups_mock.assert_called_once_with(self.context, self.node.id)
        get_volconn_mock.assert_called_once_with(self.context, self.node.id)
        get_voltgt_mock.assert_called_once_with(self.context, self.node.id)
        release_mock.assert_called_once_with(self.context, self.host,
                                             self.node.id)

    def test_no_driver(self, get_voltgt_mock, get_volconn_mock,
                       get_portgroups_mock, get_ports_mock,
                       build_driver_mock, reserve_mock, release_mock,
                       node_get_mock):
        reserve_mock.return_value = self.node
        with task_manager.TaskManager(self.context, 'fake-node-id',
                                      load_driver=False) as task:
            self.assertEqual(self.context, task.context)
            self.assertEqual(self.node, task.node)
            self.assertEqual(get_ports_mock.return_value, task.ports)
            self.assertEqual(get_portgroups_mock.return_value, task.portgroups)
            self.assertEqual(get_volconn_mock.return_value,
                             task.volume_connectors)
            self.assertEqual(get_voltgt_mock.return_value, task.volume_targets)
            self.assertIsNone(task.driver)
            self.assertFalse(task.shared)
        self.assertFalse(build_driver_mock.called)

    def test_excl_nested_acquire(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node2 = obj_utils.create_test_node(self.context,
                                           uuid=uuidutils.generate_uuid(),
                                           driver='fake')

        reserve_mock.return_value = self.node
        get_ports_mock.return_value = mock.sentinel.ports1
        get_portgroups_mock.return_value = mock.sentinel.portgroups1
        get_volconn_mock.return_value = mock.sentinel.volconn1
        get_voltgt_mock.return_value = mock.sentinel.voltgt1
        build_driver_mock.return_value = mock.sentinel.driver1

        with task_manager.TaskManager(self.context, 'node-id1') as task:
            reserve_mock.return_value = node2
            get_ports_mock.return_value = mock.sentinel.ports2
            get_portgroups_mock.return_value = mock.sentinel.portgroups2
            get_volconn_mock.return_value = mock.sentinel.volconn2
            get_voltgt_mock.return_value = mock.sentinel.voltgt2
            build_driver_mock.return_value = mock.sentinel.driver2
            with task_manager.TaskManager(self.context, 'node-id2') as task2:
                self.assertEqual(self.context, task.context)
                self.assertEqual(self.node, task.node)
                self.assertEqual(mock.sentinel.ports1, task.ports)
                self.assertEqual(mock.sentinel.portgroups1, task.portgroups)
                self.assertEqual(mock.sentinel.volconn1,
                                 task.volume_connectors)
                self.assertEqual(mock.sentinel.voltgt1, task.volume_targets)
                self.assertEqual(mock.sentinel.driver1, task.driver)
                self.assertFalse(task.shared)
                self.assertEqual(self.context, task2.context)
                self.assertEqual(node2, task2.node)
                self.assertEqual(mock.sentinel.ports2, task2.ports)
                self.assertEqual(mock.sentinel.portgroups2, task2.portgroups)
                self.assertEqual(mock.sentinel.volconn2,
                                 task2.volume_connectors)
                self.assertEqual(mock.sentinel.voltgt2, task2.volume_targets)
                self.assertEqual(mock.sentinel.driver2, task2.driver)
                self.assertFalse(task2.shared)

                self.assertEqual([mock.call(task), mock.call(task2)],
                                 build_driver_mock.call_args_list)

        self.assertEqual([mock.call(self.context, 'node-id1'),
                          mock.call(self.context, 'node-id2')],
                         node_get_mock.call_args_list)
        self.assertEqual([mock.call(self.context, self.host, 'node-id1'),
                          mock.call(self.context, self.host, 'node-id2')],
                         reserve_mock.call_args_list)
        self.assertEqual([mock.call(self.context, self.node.id),
                          mock.call(self.context, node2.id)],
                         get_ports_mock.call_args_list)
        # release should be in reverse order
        self.assertEqual([mock.call(self.context, self.host, node2.id),
                          mock.call(self.context, self.host, self.node.id)],
                         release_mock.call_args_list)

    def test_excl_lock_exception_then_lock(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        retry_attempts = 3
        self.config(node_locked_retry_attempts=retry_attempts,
                    group='conductor')

        # Fail on the first lock attempt, succeed on the second.
        reserve_mock.side_effect = [exception.NodeLocked(node='foo',
                                                         host='foo'),
                                    self.node]

        with task_manager.TaskManager(self.context, 'fake-node-id') as task:
            self.assertFalse(task.shared)

        expected_calls = [mock.call(self.context, self.host,
                                    'fake-node-id')] * 2
        reserve_mock.assert_has_calls(expected_calls)
        self.assertEqual(2, reserve_mock.call_count)

    def test_excl_lock_reserve_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        retry_attempts = 3
        self.config(node_locked_retry_attempts=retry_attempts,
                    group='conductor')
        reserve_mock.side_effect = exception.NodeLocked(node='foo',
                                                        host='foo')

        self.assertRaises(exception.NodeLocked,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id')
        node_get_mock.assert_called_with(self.context, 'fake-node-id')
        reserve_mock.assert_called_with(self.context, self.host,
                                        'fake-node-id')
        self.assertEqual(retry_attempts, reserve_mock.call_count)
        self.assertFalse(get_ports_mock.called)
        self.assertFalse(get_portgroups_mock.called)
        self.assertFalse(get_volconn_mock.called)
        self.assertFalse(get_voltgt_mock.called)
        self.assertFalse(build_driver_mock.called)
        self.assertFalse(release_mock.called)

    def test_excl_lock_get_ports_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        reserve_mock.return_value = self.node
        get_ports_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id')

        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        reserve_mock.assert_called_once_with(self.context, self.host,
                                             'fake-node-id')
        get_ports_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(build_driver_mock.called)
        release_mock.assert_called_once_with(self.context, self.host,
                                             self.node.id)

    def test_excl_lock_get_portgroups_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        reserve_mock.return_value = self.node
        get_portgroups_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id')

        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        reserve_mock.assert_called_once_with(self.context, self.host,
                                             'fake-node-id')
        get_portgroups_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(build_driver_mock.called)
        release_mock.assert_called_once_with(self.context, self.host,
                                             self.node.id)

    def test_excl_lock_get_volconn_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        reserve_mock.return_value = self.node
        get_volconn_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id')

        reserve_mock.assert_called_once_with(self.context, self.host,
                                             'fake-node-id')
        get_volconn_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(get_voltgt_mock.called)
        release_mock.assert_called_once_with(self.context, self.host,
                                             self.node.id)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')

    def test_excl_lock_get_voltgt_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        reserve_mock.return_value = self.node
        get_voltgt_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id')

        reserve_mock.assert_called_once_with(self.context, self.host,
                                             'fake-node-id')
        get_voltgt_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(build_driver_mock.called)
        release_mock.assert_called_once_with(self.context, self.host,
                                             self.node.id)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')

    def test_excl_lock_build_driver_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        reserve_mock.return_value = self.node
        build_driver_mock.side_effect = (
            exception.DriverNotFound(driver_name='foo'))

        self.assertRaises(exception.DriverNotFound,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id')

        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        reserve_mock.assert_called_once_with(self.context, self.host,
                                             'fake-node-id')
        get_ports_mock.assert_called_once_with(self.context, self.node.id)
        get_portgroups_mock.assert_called_once_with(self.context, self.node.id)
        build_driver_mock.assert_called_once_with(mock.ANY)
        release_mock.assert_called_once_with(self.context, self.host,
                                             self.node.id)

    def test_shared_lock(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.return_value = self.node
        with task_manager.TaskManager(self.context, 'fake-node-id',
                                      shared=True) as task:
            self.assertEqual(self.context, task.context)
            self.assertEqual(self.node, task.node)
            self.assertEqual(get_ports_mock.return_value, task.ports)
            self.assertEqual(get_portgroups_mock.return_value, task.portgroups)
            self.assertEqual(get_volconn_mock.return_value,
                             task.volume_connectors)
            self.assertEqual(get_voltgt_mock.return_value, task.volume_targets)
            self.assertEqual(build_driver_mock.return_value, task.driver)
            self.assertTrue(task.shared)

            build_driver_mock.assert_called_once_with(task)

        self.assertFalse(reserve_mock.called)
        self.assertFalse(release_mock.called)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        get_ports_mock.assert_called_once_with(self.context, self.node.id)
        get_portgroups_mock.assert_called_once_with(self.context, self.node.id)
        get_volconn_mock.assert_called_once_with(self.context, self.node.id)
        get_voltgt_mock.assert_called_once_with(self.context, self.node.id)

    def test_shared_lock_node_get_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.side_effect = exception.NodeNotFound(node='foo')

        self.assertRaises(exception.NodeNotFound,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id',
                          shared=True)

        self.assertFalse(reserve_mock.called)
        self.assertFalse(release_mock.called)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        self.assertFalse(get_ports_mock.called)
        self.assertFalse(get_portgroups_mock.called)
        self.assertFalse(get_volconn_mock.called)
        self.assertFalse(get_voltgt_mock.called)
        self.assertFalse(build_driver_mock.called)

    def test_shared_lock_get_ports_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.return_value = self.node
        get_ports_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id',
                          shared=True)

        self.assertFalse(reserve_mock.called)
        self.assertFalse(release_mock.called)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        get_ports_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(build_driver_mock.called)

    def test_shared_lock_get_portgroups_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.return_value = self.node
        get_portgroups_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id',
                          shared=True)

        self.assertFalse(reserve_mock.called)
        self.assertFalse(release_mock.called)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        get_portgroups_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(build_driver_mock.called)

    def test_shared_lock_get_volconn_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.return_value = self.node
        get_volconn_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id',
                          shared=True)

        self.assertFalse(reserve_mock.called)
        self.assertFalse(release_mock.called)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        get_volconn_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(get_voltgt_mock.called)

    def test_shared_lock_get_voltgt_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.return_value = self.node
        get_voltgt_mock.side_effect = exception.IronicException('foo')

        self.assertRaises(exception.IronicException,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id',
                          shared=True)

        self.assertFalse(reserve_mock.called)
        self.assertFalse(release_mock.called)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        get_voltgt_mock.assert_called_once_with(self.context, self.node.id)
        self.assertFalse(build_driver_mock.called)

    def test_shared_lock_build_driver_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.return_value = self.node
        build_driver_mock.side_effect = (
            exception.DriverNotFound(driver_name='foo'))

        self.assertRaises(exception.DriverNotFound,
                          task_manager.TaskManager,
                          self.context,
                          'fake-node-id',
                          shared=True)

        self.assertFalse(reserve_mock.called)
        self.assertFalse(release_mock.called)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        get_ports_mock.assert_called_once_with(self.context, self.node.id)
        get_portgroups_mock.assert_called_once_with(self.context, self.node.id)
        get_volconn_mock.assert_called_once_with(self.context, self.node.id)
        get_voltgt_mock.assert_called_once_with(self.context, self.node.id)
        build_driver_mock.assert_called_once_with(mock.ANY)

    def test_upgrade_lock(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        node_get_mock.return_value = self.node
        reserve_mock.return_value = self.node
        with task_manager.TaskManager(self.context, 'fake-node-id',
                                      shared=True, purpose='ham') as task:
            self.assertEqual(self.context, task.context)
            self.assertEqual(self.node, task.node)
            self.assertEqual(get_ports_mock.return_value, task.ports)
            self.assertEqual(get_portgroups_mock.return_value, task.portgroups)
            self.assertEqual(get_volconn_mock.return_value,
                             task.volume_connectors)
            self.assertEqual(get_voltgt_mock.return_value, task.volume_targets)
            self.assertEqual(build_driver_mock.return_value, task.driver)
            self.assertTrue(task.shared)
            self.assertFalse(reserve_mock.called)

            task.upgrade_lock()
            self.assertFalse(task.shared)
            self.assertEqual('ham', task._purpose)
            # second upgrade does nothing except changes the purpose
            task.upgrade_lock(purpose='spam')
            self.assertFalse(task.shared)
            self.assertEqual('spam', task._purpose)

            build_driver_mock.assert_called_once_with(mock.ANY)

        # make sure reserve() was called only once
        reserve_mock.assert_called_once_with(self.context, self.host,
                                             'fake-node-id')
        release_mock.assert_called_once_with(self.context, self.host,
                                             self.node.id)
        node_get_mock.assert_called_once_with(self.context, 'fake-node-id')
        get_ports_mock.assert_called_once_with(self.context, self.node.id)
        get_portgroups_mock.assert_called_once_with(self.context, self.node.id)
        get_volconn_mock.assert_called_once_with(self.context, self.node.id)
        get_voltgt_mock.assert_called_once_with(self.context, self.node.id)

    def test_upgrade_lock_refreshes_fsm(self, get_voltgt_mock,
                                        get_volconn_mock, get_portgroups_mock,
                                        get_ports_mock, build_driver_mock,
                                        reserve_mock, release_mock,
                                        node_get_mock):
        reserve_mock.return_value = self.node
        node_get_mock.return_value = self.node
        with task_manager.acquire(self.context, 'fake-node-id',
                                  shared=True) as task1:
            self.assertEqual(states.AVAILABLE, task1.node.provision_state)

            with task_manager.acquire(self.context, 'fake-node-id',
                                      shared=False) as task2:
                # move the node to manageable
                task2.process_event('manage')
                self.assertEqual(states.MANAGEABLE, task1.node.provision_state)

            # now upgrade our shared task and try to go to cleaning
            # this will explode if task1's FSM doesn't get refreshed
            task1.upgrade_lock()
            task1.process_event('provide')
            self.assertEqual(states.CLEANING, task1.node.provision_state)

    @mock.patch.object(task_manager.TaskManager,
                       '_notify_provision_state_change', autospec=True)
    def test_spawn_after(
            self, notify_mock, get_voltgt_mock, get_volconn_mock,
            get_portgroups_mock, get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        spawn_mock = mock.Mock(return_value=self.future_mock)
        task_release_mock = mock.Mock()
        reserve_mock.return_value = self.node

        with task_manager.TaskManager(self.context, 'node-id') as task:
            task.spawn_after(spawn_mock, 1, 2, foo='bar', cat='meow')
            task.release_resources = task_release_mock

        spawn_mock.assert_called_once_with(1, 2, foo='bar', cat='meow')
        self.future_mock.add_done_callback.assert_called_once_with(
            task._thread_release_resources)
        self.assertFalse(self.future_mock.cancel.called)
        # Since we mocked link(), we're testing that __exit__ didn't
        # release resources pending the finishing of the background
        # thread
        self.assertFalse(task_release_mock.called)
        notify_mock.assert_called_once_with(task)

    def test_spawn_after_exception_while_yielded(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        spawn_mock = mock.Mock()
        task_release_mock = mock.Mock()
        reserve_mock.return_value = self.node

        def _test_it():
            with task_manager.TaskManager(self.context, 'node-id') as task:
                task.spawn_after(spawn_mock, 1, 2, foo='bar', cat='meow')
                task.release_resources = task_release_mock
                raise exception.IronicException('foo')

        self.assertRaises(exception.IronicException, _test_it)
        self.assertFalse(spawn_mock.called)
        task_release_mock.assert_called_once_with()

    @mock.patch.object(task_manager.TaskManager,
                       '_notify_provision_state_change', autospec=True)
    def test_spawn_after_spawn_fails(
            self, notify_mock, get_voltgt_mock, get_volconn_mock,
            get_portgroups_mock, get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        spawn_mock = mock.Mock(side_effect=exception.IronicException('foo'))
        task_release_mock = mock.Mock()
        reserve_mock.return_value = self.node

        def _test_it():
            with task_manager.TaskManager(self.context, 'node-id') as task:
                task.spawn_after(spawn_mock, 1, 2, foo='bar', cat='meow')
                task.release_resources = task_release_mock

        self.assertRaises(exception.IronicException, _test_it)

        spawn_mock.assert_called_once_with(1, 2, foo='bar', cat='meow')
        task_release_mock.assert_called_once_with()
        self.assertFalse(notify_mock.called)

    def test_spawn_after_link_fails(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        self.future_mock.add_done_callback.side_effect = (
            exception.IronicException('foo'))
        spawn_mock = mock.Mock(return_value=self.future_mock)
        task_release_mock = mock.Mock()
        thr_release_mock = mock.Mock(spec_set=[])
        reserve_mock.return_value = self.node

        def _test_it():
            with task_manager.TaskManager(self.context, 'node-id') as task:
                task.spawn_after(spawn_mock, 1, 2, foo='bar', cat='meow')
                task._thread_release_resources = thr_release_mock
                task.release_resources = task_release_mock
        self.assertRaises(exception.IronicException, _test_it)

        spawn_mock.assert_called_once_with(1, 2, foo='bar', cat='meow')
        self.future_mock.add_done_callback.assert_called_once_with(
            thr_release_mock)
        self.future_mock.cancel.assert_called_once_with()
        task_release_mock.assert_called_once_with()

    def test_spawn_after_on_error_hook(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        expected_exception = exception.IronicException('foo')
        spawn_mock = mock.Mock(side_effect=expected_exception)
        task_release_mock = mock.Mock()
        on_error_handler = mock.Mock()
        reserve_mock.return_value = self.node

        def _test_it():
            with task_manager.TaskManager(self.context, 'node-id') as task:
                task.set_spawn_error_hook(on_error_handler, 'fake-argument')
                task.spawn_after(spawn_mock, 1, 2, foo='bar', cat='meow')
                task.release_resources = task_release_mock

        self.assertRaises(exception.IronicException, _test_it)

        spawn_mock.assert_called_once_with(1, 2, foo='bar', cat='meow')
        task_release_mock.assert_called_once_with()
        on_error_handler.assert_called_once_with(expected_exception,
                                                 'fake-argument')

    def test_spawn_after_on_error_hook_exception(
            self, get_voltgt_mock, get_volconn_mock, get_portgroups_mock,
            get_ports_mock, build_driver_mock,
            reserve_mock, release_mock, node_get_mock):
        expected_exception = exception.IronicException('foo')
        spawn_mock = mock.Mock(side_effect=expected_exception)
        task_release_mock = mock.Mock()
        # Raise an exception within the on_error handler
        on_error_handler = mock.Mock(side_effect=Exception('unexpected'))
        on_error_handler.__name__ = 'foo_method'
        reserve_mock.return_value = self.node

        def _test_it():
            with task_manager.TaskManager(self.context, 'node-id') as task:
                task.set_spawn_error_hook(on_error_handler, 'fake-argument')
                task.spawn_after(spawn_mock, 1, 2, foo='bar', cat='meow')
                task.release_resources = task_release_mock

        # Make sure the original exception is the one raised
        self.assertRaises(exception.IronicException, _test_it)

        spawn_mock.assert_called_once_with(1, 2, foo='bar', cat='meow')
        task_release_mock.assert_called_once_with()
        on_error_handler.assert_called_once_with(expected_exception,
                                                 'fake-argument')

    @mock.patch.object(states.machine, 'copy')
    def test_init_prepares_fsm(
            self, copy_mock, get_volconn_mock, get_voltgt_mock,
            get_portgroups_mock, get_ports_mock,
            build_driver_mock, reserve_mock, release_mock, node_get_mock):
        m = mock.Mock(spec=fsm.FSM)
        reserve_mock.return_value = self.node
        copy_mock.return_value = m
        t = task_manager.TaskManager('fake', 'fake')
        copy_mock.assert_called_once_with()
        self.assertIs(m, t.fsm)
        m.initialize.assert_called_once_with(
            start_state=self.node.provision_state,
            target_state=self.node.target_provision_state)


class TaskManagerStateModelTestCases(tests_base.TestCase):
    def setUp(self):
        super(TaskManagerStateModelTestCases, self).setUp()
        self.fsm = mock.Mock(spec=fsm.FSM)
        self.node = mock.Mock(spec=objects.Node)
        self.task = mock.Mock(spec=task_manager.TaskManager)
        self.task.fsm = self.fsm
        self.task.node = self.node

    def test_release_clears_resources(self):
        t = self.task
        t.release_resources = task_manager.TaskManager.release_resources
        t.driver = mock.Mock()
        t.ports = mock.Mock()
        t.portgroups = mock.Mock()
        t.volume_connectors = mock.Mock()
        t.volume_targets = mock.Mock()
        t.shared = True
        t._purpose = 'purpose'
        t._debug_timer = mock.Mock()
        t._debug_timer.elapsed.return_value = 3.14

        t.release_resources(t)
        self.assertIsNone(t.node)
        self.assertIsNone(t.driver)
        self.assertIsNone(t.ports)
        self.assertIsNone(t.portgroups)
        self.assertIsNone(t.volume_connectors)
        self.assertIsNone(t.volume_targets)
        self.assertIsNone(t.fsm)

    def test_process_event_fsm_raises(self):
        self.task.process_event = task_manager.TaskManager.process_event
        self.fsm.process_event.side_effect = exception.InvalidState('test')

        self.assertRaises(
            exception.InvalidState,
            self.task.process_event,
            self.task, 'fake')
        self.assertEqual(0, self.task.spawn_after.call_count)
        self.assertFalse(self.task.node.save.called)

    def test_process_event_sets_callback(self):
        cb = mock.Mock()
        arg = mock.Mock()
        kwarg = mock.Mock()
        self.task.process_event = task_manager.TaskManager.process_event
        self.task.process_event(
            self.task, 'fake', callback=cb, call_args=[arg],
            call_kwargs={'mock': kwarg})
        self.fsm.process_event.assert_called_once_with('fake',
                                                       target_state=None)
        self.task.spawn_after.assert_called_with(cb, arg, mock=kwarg)
        self.assertEqual(1, self.task.node.save.call_count)
        self.assertIsNone(self.node.last_error)

    def test_process_event_sets_callback_and_error_handler(self):
        arg = mock.Mock()
        cb = mock.Mock()
        er = mock.Mock()
        kwarg = mock.Mock()
        provision_state = 'provision_state'
        target_provision_state = 'target'
        self.node.provision_state = provision_state
        self.node.target_provision_state = target_provision_state
        self.task.process_event = task_manager.TaskManager.process_event

        self.task.process_event(
            self.task, 'fake', callback=cb, call_args=[arg],
            call_kwargs={'mock': kwarg}, err_handler=er)

        self.task.set_spawn_error_hook.assert_called_once_with(
            er, self.node, provision_state, target_provision_state)
        self.fsm.process_event.assert_called_once_with('fake',
                                                       target_state=None)
        self.task.spawn_after.assert_called_with(cb, arg, mock=kwarg)
        self.assertEqual(1, self.task.node.save.call_count)
        self.assertIsNone(self.node.last_error)
        self.assertNotEqual(provision_state, self.node.provision_state)
        self.assertNotEqual(target_provision_state,
                            self.node.target_provision_state)

    def test_process_event_sets_target_state(self):
        event = 'fake'
        tgt_state = 'target'
        provision_state = 'provision_state'
        target_provision_state = 'target_provision_state'
        self.node.provision_state = provision_state
        self.node.target_provision_state = target_provision_state
        self.task.process_event = task_manager.TaskManager.process_event
        self.task.process_event(self.task, event, target_state=tgt_state)
        self.fsm.process_event.assert_called_once_with(event,
                                                       target_state=tgt_state)
        self.assertEqual(1, self.task.node.save.call_count)
        self.assertNotEqual(provision_state, self.node.provision_state)
        self.assertNotEqual(target_provision_state,
                            self.node.target_provision_state)

    def test_process_event_callback_stable_state(self):
        callback = mock.Mock()
        for state in states.STABLE_STATES:
            self.node.provision_state = state
            self.node.target_provision_state = 'target'
            self.task.process_event = task_manager.TaskManager.process_event
            self.task.process_event(self.task, 'fake', callback=callback)
            # assert the target state is set when callback is passed
            self.assertNotEqual(states.NOSTATE,
                                self.task.node.target_provision_state)

    def test_process_event_no_callback_stable_state(self):
        for state in states.STABLE_STATES:
            self.node.provision_state = state
            self.node.target_provision_state = 'target'
            self.task.process_event = task_manager.TaskManager.process_event
            self.task.process_event(self.task, 'fake')
            # assert the target state was cleared when moving to a
            # stable state
            self.assertEqual(states.NOSTATE,
                             self.task.node.target_provision_state)

    def test_process_event_no_callback_notify(self):
        self.task.process_event = task_manager.TaskManager.process_event
        self.task.process_event(self.task, 'fake')
        self.task._notify_provision_state_change.assert_called_once_with()


@task_manager.require_exclusive_lock
def _req_excl_lock_method(*args, **kwargs):
    return (args, kwargs)


class ExclusiveLockDecoratorTestCase(tests_base.TestCase):
    def setUp(self):
        super(ExclusiveLockDecoratorTestCase, self).setUp()
        self.task = mock.Mock(spec=task_manager.TaskManager)
        self.task.context = self.context
        self.args_task_first = (self.task, 1, 2)
        self.args_task_second = (1, self.task, 2)
        self.kwargs = dict(cat='meow', dog='wuff')

    def test_with_excl_lock_task_first_arg(self):
        self.task.shared = False
        (args, kwargs) = _req_excl_lock_method(*self.args_task_first,
                                               **self.kwargs)
        self.assertEqual(self.args_task_first, args)
        self.assertEqual(self.kwargs, kwargs)

    def test_with_excl_lock_task_second_arg(self):
        self.task.shared = False
        (args, kwargs) = _req_excl_lock_method(*self.args_task_second,
                                               **self.kwargs)
        self.assertEqual(self.args_task_second, args)
        self.assertEqual(self.kwargs, kwargs)

    def test_with_shared_lock_task_first_arg(self):
        self.task.shared = True
        self.assertRaises(exception.ExclusiveLockRequired,
                          _req_excl_lock_method,
                          *self.args_task_first,
                          **self.kwargs)

    def test_with_shared_lock_task_second_arg(self):
        self.task.shared = True
        self.assertRaises(exception.ExclusiveLockRequired,
                          _req_excl_lock_method,
                          *self.args_task_second,
                          **self.kwargs)


class ThreadExceptionTestCase(tests_base.TestCase):
    def setUp(self):
        super(ThreadExceptionTestCase, self).setUp()
        self.node = mock.Mock(spec=objects.Node)
        self.node.last_error = None
        self.task = mock.Mock(spec=task_manager.TaskManager)
        self.task.node = self.node
        self.task._write_exception = task_manager.TaskManager._write_exception
        self.future_mock = mock.Mock(spec_set=['exception'])

        def async_method_foo():
            pass

        self.task._spawn_args = (async_method_foo,)

    def test_set_node_last_error(self):
        self.future_mock.exception.return_value = Exception('fiasco')
        self.task._write_exception(self.task, self.future_mock)
        self.node.save.assert_called_once_with()
        self.assertIn('fiasco', self.node.last_error)
        self.assertIn('async_method_foo', self.node.last_error)

    def test_set_node_last_error_exists(self):
        self.future_mock.exception.return_value = Exception('fiasco')
        self.node.last_error = 'oops'
        self.task._write_exception(self.task, self.future_mock)
        self.assertFalse(self.node.save.called)
        self.assertFalse(self.future_mock.exception.called)
        self.assertEqual('oops', self.node.last_error)

    def test_set_node_last_error_no_error(self):
        self.future_mock.exception.return_value = None
        self.task._write_exception(self.task, self.future_mock)
        self.assertFalse(self.node.save.called)
        self.future_mock.exception.assert_called_once_with()
        self.assertIsNone(self.node.last_error)

    @mock.patch.object(task_manager.LOG, 'exception', spec_set=True,
                       autospec=True)
    def test_set_node_last_error_cancelled(self, log_mock):
        self.future_mock.exception.side_effect = futurist.CancelledError()
        self.task._write_exception(self.task, self.future_mock)
        self.assertFalse(self.node.save.called)
        self.future_mock.exception.assert_called_once_with()
        self.assertIsNone(self.node.last_error)
        self.assertTrue(log_mock.called)


@mock.patch.object(notification_utils, 'emit_provision_set_notification',
                   autospec=True)
class ProvisionNotifyTestCase(tests_base.TestCase):
    def setUp(self):
        super(ProvisionNotifyTestCase, self).setUp()
        self.node = mock.Mock(spec=objects.Node)
        self.task = mock.Mock(spec=task_manager.TaskManager)
        self.task.node = self.node
        notifier = task_manager.TaskManager._notify_provision_state_change
        self.task.notifier = notifier
        self.task._prev_target_provision_state = 'oldtarget'
        self.task._event = 'event'

    def test_notify_no_state_change(self, emit_mock):
        self.task._event = None
        self.task.notifier(self.task)
        self.assertFalse(emit_mock.called)

    def test_notify_error_state(self, emit_mock):
        self.task._event = 'fail'
        self.task._prev_provision_state = 'fake'
        self.task.notifier(self.task)
        emit_mock.assert_called_once_with(self.task,
                                          fields.NotificationLevel.ERROR,
                                          fields.NotificationStatus.ERROR,
                                          'fake', 'oldtarget', 'fail')
        self.assertIsNone(self.task._event)

    def test_notify_unstable_to_unstable(self, emit_mock):
        self.node.provision_state = states.DEPLOYING
        self.task._prev_provision_state = states.DEPLOYWAIT
        self.task.notifier(self.task)
        emit_mock.assert_called_once_with(self.task,
                                          fields.NotificationLevel.INFO,
                                          fields.NotificationStatus.SUCCESS,
                                          states.DEPLOYWAIT,
                                          'oldtarget', 'event')

    def test_notify_stable_to_unstable(self, emit_mock):
        self.node.provision_state = states.DEPLOYING
        self.task._prev_provision_state = states.AVAILABLE
        self.task.notifier(self.task)
        emit_mock.assert_called_once_with(self.task,
                                          fields.NotificationLevel.INFO,
                                          fields.NotificationStatus.START,
                                          states.AVAILABLE,
                                          'oldtarget', 'event')

    def test_notify_unstable_to_stable(self, emit_mock):
        self.node.provision_state = states.ACTIVE
        self.task._prev_provision_state = states.DEPLOYING
        self.task.notifier(self.task)
        emit_mock.assert_called_once_with(self.task,
                                          fields.NotificationLevel.INFO,
                                          fields.NotificationStatus.END,
                                          states.DEPLOYING,
                                          'oldtarget', 'event')

    def test_notify_stable_to_stable(self, emit_mock):
        self.node.provision_state = states.MANAGEABLE
        self.task._prev_provision_state = states.AVAILABLE
        self.task.notifier(self.task)
        emit_mock.assert_called_once_with(self.task,
                                          fields.NotificationLevel.INFO,
                                          fields.NotificationStatus.SUCCESS,
                                          states.AVAILABLE,
                                          'oldtarget', 'event')

    def test_notify_resource_released(self, emit_mock):
        node = mock.Mock(spec=objects.Node)
        node.provision_state = states.DEPLOYING
        node.target_provision_state = states.ACTIVE
        task = mock.Mock(spec=task_manager.TaskManager)
        task._prev_provision_state = states.AVAILABLE
        task._prev_target_provision_state = states.NOSTATE
        task._event = 'event'
        task.node = None
        task._saved_node = node
        notifier = task_manager.TaskManager._notify_provision_state_change
        task.notifier = notifier
        task.notifier(task)
        task_arg = emit_mock.call_args[0][0]
        self.assertEqual(node, task_arg.node)
        self.assertIsNot(task, task_arg)

    def test_notify_only_once(self, emit_mock):
        self.node.provision_state = states.DEPLOYING
        self.task._prev_provision_state = states.AVAILABLE
        self.task.notifier(self.task)
        self.assertIsNone(self.task._event)
        self.task.notifier(self.task)
        self.assertEqual(1, emit_mock.call_count)
        self.assertIsNone(self.task._event)



# Copyright 2015 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Common functionalities shared between different iRMC modules.
"""
from oslo_log import log as logging
from oslo_utils import importutils
import six

from ironic.common import exception
from ironic.common.i18n import _
from ironic.conf import CONF

scci = importutils.try_import('scciclient.irmc.scci')
elcm = importutils.try_import('scciclient.irmc.elcm')

LOG = logging.getLogger(__name__)
REQUIRED_PROPERTIES = {
    'irmc_address': _("IP address or hostname of the iRMC. Required."),
    'irmc_username': _("Username for the iRMC with administrator privileges. "
                       "Required."),
    'irmc_password': _("Password for irmc_username. Required."),
}
OPTIONAL_PROPERTIES = {
    'irmc_port': _("Port to be used for iRMC operations; either 80 or 443. "
                   "The default value is 443. Optional."),
    'irmc_auth_method': _("Authentication method for iRMC operations; "
                          "either 'basic' or 'digest'. The default value is "
                          "'basic'. Optional."),
    'irmc_client_timeout': _("Timeout (in seconds) for iRMC operations. "
                             "The default value is 60. Optional."),
    'irmc_sensor_method': _("Sensor data retrieval method; either "
                            "'ipmitool' or 'scci'. The default value is "
                            "'ipmitool'. Optional."),
    'irmc_snmp_version': _("SNMP protocol version; either 'v1', 'v2c', or "
                           "'v3'. The default value is 'v2c'. Optional."),
    'irmc_snmp_port': _("SNMP port. The default is 161. Optional."),
    'irmc_snmp_community': _("SNMP community required for versions 'v1' and "
                             "'v2c'. The default value is 'public'. "
                             "Optional."),
    'irmc_snmp_security': _("SNMP security name required for version 'v3'. "
                            "Optional."),
}

COMMON_PROPERTIES = REQUIRED_PROPERTIES.copy()
COMMON_PROPERTIES.update(OPTIONAL_PROPERTIES)


def parse_driver_info(node):
    """Gets the specific Node driver info.

    This method validates whether the 'driver_info' property of the
    supplied node contains the required information for this driver.

    :param node: An ironic node object.
    :returns: A dict containing information from driver_info
        and default values.
    :raises: InvalidParameterValue if invalid value is contained
        in the 'driver_info' property.
    :raises: MissingParameterValue if some mandatory key is missing
        in the 'driver_info' property.
    """
    info = node.driver_info
    missing_info = [key for key in REQUIRED_PROPERTIES if not info.get(key)]
    if missing_info:
        raise exception.MissingParameterValue(_(
            "Missing the following iRMC parameters in node's"
            " driver_info: %s.") % missing_info)

    req = {key: value for key, value in info.items()
           if key in REQUIRED_PROPERTIES}
    # corresponding config names don't have 'irmc_' prefix
    opt = {param: info.get(param, CONF.irmc.get(param[len('irmc_'):]))
           for param in OPTIONAL_PROPERTIES}
    d_info = dict(req, **opt)

    error_msgs = []
    if (d_info['irmc_auth_method'].lower() not in ('basic', 'digest')):
        error_msgs.append(
            _("Value '%s' is not supported for 'irmc_auth_method'.") %
            d_info['irmc_auth_method'])
    if d_info['irmc_port'] not in (80, 443):
        error_msgs.append(
            _("Value '%s' is not supported for 'irmc_port'.") %
            d_info['irmc_port'])
    if not isinstance(d_info['irmc_client_timeout'], int):
        error_msgs.append(
            _("Value '%s' is not an integer for 'irmc_client_timeout'") %
            d_info['irmc_client_timeout'])
    if d_info['irmc_sensor_method'].lower() not in ('ipmitool', 'scci'):
        error_msgs.append(
            _("Value '%s' is not supported for 'irmc_sensor_method'.") %
            d_info['irmc_sensor_method'])
    if d_info['irmc_snmp_version'].lower() not in ('v1', 'v2c', 'v3'):
        error_msgs.append(
            _("Value '%s' is not supported for 'irmc_snmp_version'.") %
            d_info['irmc_snmp_version'])
    if not isinstance(d_info['irmc_snmp_port'], int):
        error_msgs.append(
            _("Value '%s' is not an integer for 'irmc_snmp_port'") %
            d_info['irmc_snmp_port'])
    if (d_info['irmc_snmp_version'].lower() in ('v1', 'v2c') and
        d_info['irmc_snmp_community'] and
        not isinstance(d_info['irmc_snmp_community'], six.string_types)):
        error_msgs.append(
            _("Value '%s' is not a string for 'irmc_snmp_community'") %
            d_info['irmc_snmp_community'])
    if d_info['irmc_snmp_version'].lower() == 'v3':
        if d_info['irmc_snmp_security']:
            if not isinstance(d_info['irmc_snmp_security'], six.string_types):
                error_msgs.append(
                    _("Value '%s' is not a string for "
                      "'irmc_snmp_security'") % d_info['irmc_snmp_security'])
        else:
            error_msgs.append(
                _("'irmc_snmp_security' has to be set for SNMP version 3."))
    if error_msgs:
        msg = (_("The following errors were encountered while parsing "
                 "driver_info:\n%s") % "\n".join(error_msgs))
        raise exception.InvalidParameterValue(msg)

    return d_info


def get_irmc_client(node):
    """Gets an iRMC SCCI client.

    Given an ironic node object, this method gives back a iRMC SCCI client
    to do operations on the iRMC.

    :param node: An ironic node object.
    :returns: scci_cmd partial function which takes a SCCI command param.
    :raises: InvalidParameterValue on invalid inputs.
    :raises: MissingParameterValue if some mandatory information
        is missing on the node
    """
    driver_info = parse_driver_info(node)

    scci_client = scci.get_client(
        driver_info['irmc_address'],
        driver_info['irmc_username'],
        driver_info['irmc_password'],
        port=driver_info['irmc_port'],
        auth_method=driver_info['irmc_auth_method'],
        client_timeout=driver_info['irmc_client_timeout'])
    return scci_client


def update_ipmi_properties(task):
    """Update ipmi properties to node driver_info.

    :param task: A task from TaskManager.
    """
    node = task.node
    info = node.driver_info

    # updating ipmi credentials
    info['ipmi_address'] = info.get('irmc_address')
    info['ipmi_username'] = info.get('irmc_username')
    info['ipmi_password'] = info.get('irmc_password')

    # saving ipmi credentials to task object
    task.node.driver_info = info


def get_irmc_report(node):
    """Gets iRMC SCCI report.

    Given an ironic node object, this method gives back a iRMC SCCI report.

    :param node: An ironic node object.
    :returns: A xml.etree.ElementTree object.
    :raises: InvalidParameterValue on invalid inputs.
    :raises: MissingParameterValue if some mandatory information
        is missing on the node.
    :raises: scci.SCCIInvalidInputError if required parameters are invalid.
    :raises: scci.SCCIClientError if SCCI failed.
    """
    driver_info = parse_driver_info(node)

    return scci.get_report(
        driver_info['irmc_address'],
        driver_info['irmc_username'],
        driver_info['irmc_password'],
        port=driver_info['irmc_port'],
        auth_method=driver_info['irmc_auth_method'],
        client_timeout=driver_info['irmc_client_timeout'])


def set_secure_boot_mode(node, enable):
    """Enable or disable UEFI Secure Boot

    :param node: An ironic node object.
    :param enable: Boolean value. True if the secure boot to be
        enabled.
    :raises: IRMCOperationError if the operation fails.
    """
    driver_info = parse_driver_info(node)

    try:
        elcm.set_secure_boot_mode(driver_info, enable)
        LOG.info("Set secure boot to %(flag)s for node %(node)s",
                 {'flag': enable, 'node': node.uuid})
    except scci.SCCIError as irmc_exception:
        LOG.error("Failed to set secure boot to %(flag)s for node %(node)s",
                  {'flag': enable, 'node': node.uuid})
        raise exception.IRMCOperationError(
            operation=_("setting secure boot mode"),
            error=irmc_exception)



# Copyright (2015-2017) Hewlett Packard Enterprise Development LP
# Copyright (2015-2017) Universidade Federal de Campina Grande
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_utils import importutils
from oslo_utils import uuidutils

from ironic.common import driver_factory
from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.oneview import common
from ironic.drivers.modules.oneview import deploy_utils
from ironic.drivers.modules.oneview import management
from ironic.drivers.modules.oneview import power
from ironic.tests.unit.conductor import mgr_utils
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils

client_exception = importutils.try_import('hpOneView.exceptions')


class OneViewPowerDriverTestCase(db_base.DbTestCase):

    def setUp(self):
        super(OneViewPowerDriverTestCase, self).setUp()
        self.config(manager_url='https://1.2.3.4', group='oneview')
        self.config(username='user', group='oneview')
        self.config(password='password', group='oneview')
        self.config(tls_cacert_file='ca_file', group='oneview')
        self.config(allow_insecure_connections=False, group='oneview')

        mgr_utils.mock_the_extension_manager(driver='fake_oneview')
        self.driver = driver_factory.get_driver('fake_oneview')

        self.node = obj_utils.create_test_node(
            self.context, driver='fake_oneview',
            properties=db_utils.get_test_oneview_properties(),
            driver_info=db_utils.get_test_oneview_driver_info(),
        )
        self.info = common.get_oneview_info(self.node)
        deploy_utils.is_node_in_use_by_oneview = mock.Mock(return_value=False)

    @mock.patch.object(common, 'validate_oneview_resources_compatibility')
    def test_validate(self, mock_validate):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.driver.power.validate(task)
            self.assertTrue(mock_validate.called)

    def test_validate_missing_parameter(self):
        node = obj_utils.create_test_node(
            self.context, uuid=uuidutils.generate_uuid(),
            id=999, driver='fake_oneview')
        with task_manager.acquire(self.context, node.uuid) as task:
            self.assertRaises(
                exception.MissingParameterValue,
                task.driver.power.validate,
                task)

    @mock.patch.object(common, 'validate_oneview_resources_compatibility')
    def test_validate_exception(self, mock_validate):
        mock_validate.side_effect = exception.OneViewError('message')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.InvalidParameterValue,
                task.driver.power.validate,
                task)

    def test_validate_node_in_use_by_oneview(self):
        deploy_utils.is_node_in_use_by_oneview.return_value = True
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.InvalidParameterValue,
                task.driver.power.validate,
                task)

    def test_get_properties(self):
        expected = common.COMMON_PROPERTIES
        self.assertEqual(expected, self.driver.power.get_properties())

    @mock.patch.object(common, 'get_hponeview_client')
    def test_get_power_state(self, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = {'powerState': 'On'}
        client.server_hardware.get.return_value = server_hardware
        with task_manager.acquire(self.context, self.node.uuid) as task:
            power_state = self.driver.power.get_power_state(task)
            self.assertEqual(states.POWER_ON, power_state)

    @mock.patch.object(common, 'get_hponeview_client')
    def test_get_power_state_fail(self, mock_get_ov_client):
        client = mock_get_ov_client()
        exc = client_exception.HPOneViewException()
        client.server_hardware.get.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.OneViewError,
                self.driver.power.get_power_state,
                task)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_on(self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.set_power_state(task, states.POWER_ON)
            self.assertTrue(mock_set_boot_device.called)
            update = client.server_hardware.update_power_state
            update.assert_called_once_with(power.POWER_ON, server_hardware,
                                           timeout=-1)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_on_with_timeout(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.set_power_state(task, states.POWER_ON, timeout=2)
            self.assertTrue(mock_set_boot_device.called)
            update = client.server_hardware.update_power_state
            update.assert_called_once_with(power.POWER_ON, server_hardware,
                                           timeout=2)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_off(self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.set_power_state(task, states.POWER_OFF)
            self.assertFalse(mock_set_boot_device.called)
            update = client.server_hardware.update_power_state
            update.assert_called_once_with(power.POWER_OFF, server_hardware,
                                           timeout=-1)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_off_with_timeout(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.set_power_state(task, states.POWER_OFF,
                                              timeout=2)
            self.assertFalse(mock_set_boot_device.called)
            update = client.server_hardware.update_power_state
            update.assert_called_once_with(power.POWER_OFF, server_hardware,
                                           timeout=2)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_reboot(self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.set_power_state(task, states.REBOOT)
            calls = [mock.call(power.POWER_OFF, server_hardware, timeout=-1),
                     mock.call(power.POWER_ON, server_hardware, timeout=-1)]
            update = client.server_hardware.update_power_state
            update.assert_has_calls(calls)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_soft_reboot(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        self.driver.power.client = client
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.set_power_state(task, states.SOFT_REBOOT)
            calls = [mock.call(power.SOFT_POWER_OFF, server_hardware,
                               timeout=-1),
                     mock.call(power.POWER_ON, server_hardware, timeout=-1)]
            update = client.server_hardware.update_power_state
            update.assert_has_calls(calls)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_on_fail(self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        exc = client_exception.HPOneViewException()
        client.server_hardware.update_power_state.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.OneViewError,
                self.driver.power.set_power_state,
                task,
                states.POWER_ON)
            mock_set_boot_device.assert_called_once_with(task)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_off_fail(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        exc = client_exception.HPOneViewException()
        client.server_hardware.update_power_state.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.OneViewError,
                self.driver.power.set_power_state,
                task,
                states.POWER_OFF)
            self.assertFalse(mock_set_boot_device.called)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_reboot_fail_with_hardware_on(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = {'powerState': 'On'}
        client.server_hardware.get.return_value = server_hardware
        exc = client_exception.HPOneViewException()
        client.server_hardware.update_power_state.side_effect = exc
        self.driver.power.client = client
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.OneViewError,
                self.driver.power.reboot,
                task)
            self.assertFalse(mock_set_boot_device.called)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_reboot_fail_with_hardware_off(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = {'powerState': 'Off'}
        client.server_hardware.get.return_value = server_hardware
        exc = client_exception.HPOneViewException()
        client.server_hardware.update_power_state.side_effect = exc
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.OneViewError,
                self.driver.power.reboot,
                task)
            mock_set_boot_device.assert_called_once_with(task)

    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_invalid_state(self, mock_set_boot_device):
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.InvalidParameterValue,
                self.driver.power.set_power_state,
                task,
                'fake_state')
            self.assertFalse(mock_set_boot_device.called)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_reboot_with_hardware_on(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = {'powerState': 'On'}
        client.server_hardware.get.return_value = server_hardware
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.reboot(task)
            calls = [mock.call(power.POWER_OFF, server_hardware, timeout=-1),
                     mock.call(power.POWER_ON, server_hardware, timeout=-1)]
            update = client.server_hardware.update_power_state
            update.assert_has_calls(calls)
            mock_set_boot_device.assert_called_once_with(task)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_reboot_with_hardware_off(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = {'powerState': 'Off'}
        client.server_hardware.get.return_value = server_hardware
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.reboot(task, timeout=-1)
            update = client.server_hardware.update_power_state
            update.assert_called_once_with(power.POWER_ON, server_hardware,
                                           timeout=-1)
            mock_set_boot_device.assert_called_once_with(task)

    @mock.patch.object(common, 'get_hponeview_client')
    @mock.patch.object(management, 'set_boot_device')
    def test_set_power_reboot_with_hardware_off_with_timeout(
            self, mock_set_boot_device, mock_get_ov_client):
        client = mock_get_ov_client()
        server_hardware = {'powerState': 'Off'}
        client.server_hardware.get.return_value = server_hardware
        server_hardware = self.node.driver_info.get('server_hardware_uri')
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.driver.power.reboot(task, timeout=2)
            update = client.server_hardware.update_power_state
            update.assert_called_once_with(power.POWER_ON, server_hardware,
                                           timeout=2)
            mock_set_boot_device.assert_called_once_with(task)

    def test_get_supported_power_states(self):
        with task_manager.acquire(self.context, self.node.uuid,
                                  shared=True) as task:
            supported_power_states = (
                task.driver.power.get_supported_power_states(task))
            self.assertEqual(set(power.SET_POWER_STATE_MAP),
                             set(supported_power_states))



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.api.controllers.v1 import versions
from ironic.tests.unit.api import base


class TestRoot(base.BaseApiTest):

    def test_get_root(self):
        response = self.get_json('/', path_prefix='')
        # Check fields are not empty
        [self.assertNotIn(f, ['', []]) for f in response]

        self.assertEqual('OpenStack Ironic API', response['name'])
        self.assertTrue(response['description'])
        self.assertEqual([response['default_version']], response['versions'])

        version1 = response['default_version']
        self.assertEqual('v1', version1['id'])
        self.assertEqual('CURRENT', version1['status'])
        self.assertEqual(versions.min_version_string(),
                         version1['min_version'])
        self.assertEqual(versions.max_version_string(), version1['version'])


class TestV1Root(base.BaseApiTest):

    def _test_get_root(self, headers=None, additional_expected_resources=None):
        if headers is None:
            headers = {}
        if additional_expected_resources is None:
            additional_expected_resources = []
        data = self.get_json('/', headers=headers)
        self.assertEqual('v1', data['id'])
        # Check fields are not empty
        for f in data:
            self.assertNotIn(f, ['', []])
        # Check if all known resources are present and there are no extra ones.
        not_resources = ('id', 'links', 'media_types')
        actual_resources = tuple(set(data.keys()) - set(not_resources))
        expected_resources = (['chassis', 'drivers', 'nodes', 'ports'] +
                              additional_expected_resources)
        self.assertEqual(sorted(expected_resources), sorted(actual_resources))
        self.assertIn({'type': 'application/vnd.openstack.ironic.v1+json',
                       'base': 'application/json'}, data['media_types'])

    def test_get_v1_root(self):
        self._test_get_root()

    def test_get_v1_22_root(self):
        self._test_get_root(headers={'X-OpenStack-Ironic-API-Version': '1.22'},
                            additional_expected_resources=['heartbeat',
                                                           'lookup'])

    def test_get_v1_23_root(self):
        self._test_get_root(headers={'X-OpenStack-Ironic-API-Version': '1.23'},
                            additional_expected_resources=['heartbeat',
                                                           'lookup',
                                                           'portgroups'])

    def test_get_v1_32_root(self):
        self._test_get_root(headers={'X-OpenStack-Ironic-API-Version': '1.32'},
                            additional_expected_resources=['heartbeat',
                                                           'lookup',
                                                           'portgroups',
                                                           'volume'])



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

import mock
from oslo_config import cfg
from six.moves import http_client
from testtools import matchers

from ironic.api.controllers import base as api_base
from ironic.api.controllers.v1 import driver
from ironic.api.controllers.v1 import versions as api_versions
from ironic.common import exception
from ironic.conductor import rpcapi
from ironic.drivers import base as driver_base
from ironic.tests.unit.api import base


class TestListDrivers(base.BaseApiTest):
    d1 = 'fake-driver1'
    d2 = 'fake-driver2'
    d3 = 'fake-hardware-type'
    h1 = 'fake-host1'
    h2 = 'fake-host2'

    def register_fake_conductors(self):
        c1 = self.dbapi.register_conductor({
            'hostname': self.h1,
            'drivers': [self.d1, self.d2],
        })
        c2 = self.dbapi.register_conductor({
            'hostname': self.h2,
            'drivers': [self.d2],
        })
        for c in (c1, c2):
            self.dbapi.register_conductor_hardware_interfaces(
                c.id, self.d3, 'deploy', ['iscsi', 'direct'], 'direct')

    def _test_drivers(self, use_dynamic, detail=False, latest_if=False):
        self.register_fake_conductors()
        headers = {}
        expected = [
            {'name': self.d1, 'hosts': [self.h1], 'type': 'classic'},
            {'name': self.d2, 'hosts': [self.h1, self.h2], 'type': 'classic'},
            {'name': self.d3, 'hosts': [self.h1, self.h2], 'type': 'dynamic'},
        ]
        expected = sorted(expected, key=lambda d: d['name'])
        if use_dynamic:
            if latest_if:
                headers[api_base.Version.string] = \
                    api_versions.max_version_string()
            else:
                headers[api_base.Version.string] = '1.30'

        path = '/drivers'
        if detail:
            path += '?detail=True'
        data = self.get_json(path, headers=headers)

        self.assertEqual(len(expected), len(data['drivers']))
        drivers = sorted(data['drivers'], key=lambda d: d['name'])
        for i in range(len(expected)):
            d = drivers[i]
            e = expected[i]

            self.assertEqual(e['name'], d['name'])
            self.assertEqual(sorted(e['hosts']), sorted(d['hosts']))
            self.validate_link(d['links'][0]['href'])
            self.validate_link(d['links'][1]['href'])

            if use_dynamic:
                self.assertEqual(e['type'], d['type'])

            # NOTE(jroll) we don't test detail=True with use_dynamic=False
            # as this case can't actually happen.
            if detail:
                self.assertIn('default_deploy_interface', d)
                if latest_if:
                    self.assertIn('default_rescue_interface', d)
                    self.assertIn('enabled_rescue_interfaces', d)
                    self.assertIn('default_storage_interface', d)
                    self.assertIn('enabled_storage_interfaces', d)
                else:
                    self.assertNotIn('default_rescue_interface', d)
                    self.assertNotIn('enabled_rescue_interfaces', d)
                    self.assertNotIn('default_storage_interface', d)
                    self.assertNotIn('enabled_storage_interfaces', d)
            else:
                # ensure we don't spill these fields into driver listing
                # one should be enough
                self.assertNotIn('default_deploy_interface', d)

    def test_drivers(self):
        self._test_drivers(False)

    def test_drivers_with_dynamic(self):
        self._test_drivers(True)

    def _test_drivers_with_dynamic_detailed(self, latest_if=False):
        with mock.patch.object(self.dbapi, 'list_hardware_type_interfaces',
                               autospec=True) as mock_hw:
            mock_hw.return_value = [
                {
                    'hardware_type': self.d3,
                    'interface_type': 'deploy',
                    'interface_name': 'iscsi',
                    'default': False,
                },
                {
                    'hardware_type': self.d3,
                    'interface_type': 'deploy',
                    'interface_name': 'direct',
                    'default': True,
                },
            ]

            self._test_drivers(True, detail=True, latest_if=latest_if)

    def test_drivers_with_dynamic_detailed(self):
        self._test_drivers_with_dynamic_detailed()

    def test_drivers_with_dynamic_detailed_storage_interface(self):
        self._test_drivers_with_dynamic_detailed(latest_if=True)

    def _test_drivers_type_filter(self, requested_type):
        self.register_fake_conductors()
        headers = {api_base.Version.string: '1.30'}
        data = self.get_json('/drivers?type=%s' % requested_type,
                             headers=headers)
        for d in data['drivers']:
            # just check it's the right type, other tests handle the rest
            self.assertEqual(requested_type, d['type'])

    def test_drivers_type_filter_classic(self):
        self._test_drivers_type_filter('classic')

    def test_drivers_type_filter_dynamic(self):
        self._test_drivers_type_filter('dynamic')

    def test_drivers_type_filter_bad_version(self):
        headers = {api_base.Version.string: '1.29'}
        data = self.get_json('/drivers?type=classic',
                             headers=headers,
                             expect_errors=True)
        self.assertEqual(http_client.NOT_ACCEPTABLE, data.status_code)

    def test_drivers_type_filter_bad_value(self):
        headers = {api_base.Version.string: '1.30'}
        data = self.get_json('/drivers?type=working',
                             headers=headers,
                             expect_errors=True)
        self.assertEqual(http_client.BAD_REQUEST, data.status_code)

    def test_drivers_detail_bad_version(self):
        headers = {api_base.Version.string: '1.29'}
        data = self.get_json('/drivers?detail=True',
                             headers=headers,
                             expect_errors=True)
        self.assertEqual(http_client.NOT_ACCEPTABLE, data.status_code)

    def test_drivers_detail_bad_version_false(self):
        headers = {api_base.Version.string: '1.29'}
        data = self.get_json('/drivers?detail=False',
                             headers=headers,
                             expect_errors=True)
        self.assertEqual(http_client.NOT_ACCEPTABLE, data.status_code)

    def test_drivers_no_active_conductor(self):
        data = self.get_json('/drivers')
        self.assertThat(data['drivers'], matchers.HasLength(0))
        self.assertEqual([], data['drivers'])

    @mock.patch.object(rpcapi.ConductorAPI, 'get_driver_properties')
    def _test_drivers_get_one_ok(self, use_dynamic, mock_driver_properties,
                                 latest_if=False):
        # get_driver_properties mock is required by validate_link()
        self.register_fake_conductors()

        if use_dynamic:
            driver = self.d3
            driver_type = 'dynamic'
            hosts = [self.h1, self.h2]
        else:
            driver = self.d1
            driver_type = 'classic'
            hosts = [self.h1]

        headers = {}
        if latest_if:
            headers[api_base.Version.string] = \
                api_versions.max_version_string()
        else:
            headers[api_base.Version.string] = '1.30'

        data = self.get_json('/drivers/%s' % driver,
                             headers=headers)

        self.assertEqual(driver, data['name'])
        self.assertEqual(sorted(hosts), sorted(data['hosts']))
        self.assertIn('properties', data)
        self.assertEqual(driver_type, data['type'])

        if use_dynamic:
            for iface in driver_base.ALL_INTERFACES:
                if latest_if or iface not in ['rescue', 'storage']:
                    self.assertIn('default_%s_interface' % iface, data)
                    self.assertIn('enabled_%s_interfaces' % iface, data)
            self.assertIsNotNone(data['default_deploy_interface'])
            self.assertIsNotNone(data['enabled_deploy_interfaces'])
        else:
            self.assertIsNone(data['default_deploy_interface'])
            self.assertIsNone(data['enabled_deploy_interfaces'])

        self.validate_link(data['links'][0]['href'])
        self.validate_link(data['links'][1]['href'])
        self.validate_link(data['properties'][0]['href'])
        self.validate_link(data['properties'][1]['href'])

    def test_drivers_get_one_ok_classic(self):
        self._test_drivers_get_one_ok(False)

    def _test_drivers_get_one_ok_dynamic(self, latest_if=False):
        with mock.patch.object(self.dbapi, 'list_hardware_type_interfaces',
                               autospec=True) as mock_hw:
            mock_hw.return_value = [
                {
                    'hardware_type': self.d3,
                    'interface_type': 'deploy',
                    'interface_name': 'iscsi',
                    'default': False,
                },
                {
                    'hardware_type': self.d3,
                    'interface_type': 'deploy',
                    'interface_name': 'direct',
                    'default': True,
                },
            ]

            self._test_drivers_get_one_ok(True, latest_if=latest_if)
            mock_hw.assert_called_once_with([self.d3])

    def test_drivers_get_one_ok_dynamic_base_interfaces(self):
        self._test_drivers_get_one_ok_dynamic()

    def test_drivers_get_one_ok_dynamic_latest_interfaces(self):
        self._test_drivers_get_one_ok_dynamic(latest_if=True)

    def test_driver_properties_hidden_in_lower_version(self):
        self.register_fake_conductors()
        data = self.get_json('/drivers/%s' % self.d1,
                             headers={api_base.Version.string: '1.8'})
        self.assertNotIn('properties', data)

    def test_driver_type_hidden_in_lower_version(self):
        self.register_fake_conductors()
        data = self.get_json('/drivers/%s' % self.d1,
                             headers={api_base.Version.string: '1.14'})
        self.assertNotIn('type', data)

    def test_drivers_get_one_not_found(self):
        response = self.get_json('/drivers/%s' % self.d1, expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, response.status_int)

    def _test_links(self, public_url=None):
        cfg.CONF.set_override('public_endpoint', public_url, 'api')
        self.register_fake_conductors()
        data = self.get_json('/drivers/%s' % self.d1)
        self.assertIn('links', data)
        self.assertEqual(2, len(data['links']))
        self.assertIn(self.d1, data['links'][0]['href'])
        for l in data['links']:
            bookmark = l['rel'] == 'bookmark'
            self.assertTrue(self.validate_link(l['href'], bookmark=bookmark))

        if public_url is not None:
            expected = [{'href': '%s/v1/drivers/%s' % (public_url, self.d1),
                         'rel': 'self'},
                        {'href': '%s/drivers/%s' % (public_url, self.d1),
                         'rel': 'bookmark'}]
            for i in expected:
                self.assertIn(i, data['links'])

    def test_links(self):
        self._test_links()

    def test_links_public_url(self):
        self._test_links(public_url='http://foo')

    @mock.patch.object(rpcapi.ConductorAPI, 'driver_vendor_passthru')
    def test_driver_vendor_passthru_sync(self, mocked_driver_vendor_passthru):
        self.register_fake_conductors()
        mocked_driver_vendor_passthru.return_value = {
            'return': {'return_key': 'return_value'},
            'async': False,
            'attach': False}
        response = self.post_json(
            '/drivers/%s/vendor_passthru/do_test' % self.d1,
            {'test_key': 'test_value'})
        self.assertEqual(http_client.OK, response.status_int)
        self.assertEqual(mocked_driver_vendor_passthru.return_value['return'],
                         response.json)

    @mock.patch.object(rpcapi.ConductorAPI, 'driver_vendor_passthru')
    def test_driver_vendor_passthru_async(self, mocked_driver_vendor_passthru):
        self.register_fake_conductors()
        mocked_driver_vendor_passthru.return_value = {'return': None,
                                                      'async': True,
                                                      'attach': False}
        response = self.post_json(
            '/drivers/%s/vendor_passthru/do_test' % self.d1,
            {'test_key': 'test_value'})
        self.assertEqual(http_client.ACCEPTED, response.status_int)
        self.assertIsNone(mocked_driver_vendor_passthru.return_value['return'])

    @mock.patch.object(rpcapi.ConductorAPI, 'driver_vendor_passthru')
    def test_driver_vendor_passthru_put(self, mocked_driver_vendor_passthru):
        self.register_fake_conductors()
        return_value = {'return': None, 'async': True, 'attach': False}
        mocked_driver_vendor_passthru.return_value = return_value
        response = self.put_json(
            '/drivers/%s/vendor_passthru/do_test' % self.d1,
            {'test_key': 'test_value'})
        self.assertEqual(http_client.ACCEPTED, response.status_int)
        self.assertEqual(return_value['return'], response.json)

    @mock.patch.object(rpcapi.ConductorAPI, 'driver_vendor_passthru')
    def test_driver_vendor_passthru_get(self, mocked_driver_vendor_passthru):
        self.register_fake_conductors()
        return_value = {'return': 'foo', 'async': False, 'attach': False}
        mocked_driver_vendor_passthru.return_value = return_value
        response = self.get_json(
            '/drivers/%s/vendor_passthru/do_test' % self.d1)
        self.assertEqual(return_value['return'], response)

    @mock.patch.object(rpcapi.ConductorAPI, 'driver_vendor_passthru')
    def test_driver_vendor_passthru_delete(self, mock_driver_vendor_passthru):
        self.register_fake_conductors()
        return_value = {'return': None, 'async': True, 'attach': False}
        mock_driver_vendor_passthru.return_value = return_value
        response = self.delete(
            '/drivers/%s/vendor_passthru/do_test' % self.d1)
        self.assertEqual(http_client.ACCEPTED, response.status_int)
        self.assertEqual(return_value['return'], response.json)

    def test_driver_vendor_passthru_driver_not_found(self):
        # tests when given driver is not found
        # e.g. get_topic_for_driver fails to find the driver
        response = self.post_json(
            '/drivers/%s/vendor_passthru/do_test' % self.d1,
            {'test_key': 'test_value'},
            expect_errors=True)

        self.assertEqual(http_client.NOT_FOUND, response.status_int)

    def test_driver_vendor_passthru_method_not_found(self):
        response = self.post_json(
            '/drivers/%s/vendor_passthru' % self.d1,
            {'test_key': 'test_value'},
            expect_errors=True)

        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        error = json.loads(response.json['error_message'])
        self.assertEqual('Missing argument: "method"',
                         error['faultstring'])

    @mock.patch.object(rpcapi.ConductorAPI,
                       'get_driver_vendor_passthru_methods')
    def test_driver_vendor_passthru_methods(self, get_methods_mock):
        self.register_fake_conductors()
        return_value = {'foo': 'bar'}
        get_methods_mock.return_value = return_value
        path = '/drivers/%s/vendor_passthru/methods' % self.d1

        data = self.get_json(path)
        self.assertEqual(return_value, data)
        get_methods_mock.assert_called_once_with(mock.ANY, self.d1,
                                                 topic=mock.ANY)

        # Now let's test the cache: Reset the mock
        get_methods_mock.reset_mock()

        # Call it again
        data = self.get_json(path)
        self.assertEqual(return_value, data)
        # Assert RPC method wasn't called this time
        self.assertFalse(get_methods_mock.called)

    @mock.patch.object(rpcapi.ConductorAPI, 'get_raid_logical_disk_properties')
    def test_raid_logical_disk_properties(self, disk_prop_mock):
        driver._RAID_PROPERTIES = {}
        self.register_fake_conductors()
        properties = {'foo': 'description of foo'}
        disk_prop_mock.return_value = properties
        path = '/drivers/%s/raid/logical_disk_properties' % self.d1
        data = self.get_json(path,
                             headers={api_base.Version.string: "1.12"})
        self.assertEqual(properties, data)
        disk_prop_mock.assert_called_once_with(mock.ANY, self.d1,
                                               topic=mock.ANY)

    @mock.patch.object(rpcapi.ConductorAPI, 'get_raid_logical_disk_properties')
    def test_raid_logical_disk_properties_older_version(self, disk_prop_mock):
        driver._RAID_PROPERTIES = {}
        self.register_fake_conductors()
        properties = {'foo': 'description of foo'}
        disk_prop_mock.return_value = properties
        path = '/drivers/%s/raid/logical_disk_properties' % self.d1
        ret = self.get_json(path,
                            headers={api_base.Version.string: "1.4"},
                            expect_errors=True)
        self.assertEqual(http_client.NOT_ACCEPTABLE, ret.status_code)

    @mock.patch.object(rpcapi.ConductorAPI, 'get_raid_logical_disk_properties')
    def test_raid_logical_disk_properties_cached(self, disk_prop_mock):
        # only one RPC-conductor call will be made and the info cached
        # for subsequent requests
        driver._RAID_PROPERTIES = {}
        self.register_fake_conductors()
        properties = {'foo': 'description of foo'}
        disk_prop_mock.return_value = properties
        path = '/drivers/%s/raid/logical_disk_properties' % self.d1
        for i in range(3):
            data = self.get_json(path,
                                 headers={api_base.Version.string: "1.12"})
            self.assertEqual(properties, data)
        disk_prop_mock.assert_called_once_with(mock.ANY, self.d1,
                                               topic=mock.ANY)
        self.assertEqual(properties, driver._RAID_PROPERTIES[self.d1])

    @mock.patch.object(rpcapi.ConductorAPI, 'get_raid_logical_disk_properties')
    def test_raid_logical_disk_properties_iface_not_supported(
            self, disk_prop_mock):
        driver._RAID_PROPERTIES = {}
        self.register_fake_conductors()
        disk_prop_mock.side_effect = exception.UnsupportedDriverExtension(
            extension='raid', driver='fake')
        path = '/drivers/%s/raid/logical_disk_properties' % self.d1
        ret = self.get_json(path,
                            headers={api_base.Version.string: "1.12"},
                            expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, ret.status_code)
        self.assertTrue(ret.json['error_message'])
        disk_prop_mock.assert_called_once_with(mock.ANY, self.d1,
                                               topic=mock.ANY)


@mock.patch.object(rpcapi.ConductorAPI, 'get_driver_properties')
@mock.patch.object(rpcapi.ConductorAPI, 'get_topic_for_driver')
class TestDriverProperties(base.BaseApiTest):

    def test_driver_properties_fake(self, mock_topic, mock_properties):
        # Can get driver properties for fake driver.
        driver._DRIVER_PROPERTIES = {}
        driver_name = 'fake'
        mock_topic.return_value = 'fake_topic'
        mock_properties.return_value = {'prop1': 'Property 1. Required.'}
        data = self.get_json('/drivers/%s/properties' % driver_name)
        self.assertEqual(mock_properties.return_value, data)
        mock_topic.assert_called_once_with(driver_name)
        mock_properties.assert_called_once_with(mock.ANY, driver_name,
                                                topic=mock_topic.return_value)
        self.assertEqual(mock_properties.return_value,
                         driver._DRIVER_PROPERTIES[driver_name])

    def test_driver_properties_hw_type(self, mock_topic, mock_properties):
        # Can get driver properties for manual-management hardware type
        driver._DRIVER_PROPERTIES = {}
        driver_name = 'manual-management'
        mock_topic.return_value = 'fake_topic'
        mock_properties.return_value = {'prop1': 'Property 1. Required.'}

        with mock.patch.object(self.dbapi, 'get_active_hardware_type_dict',
                               autospec=True) as mock_hw_type:
            mock_hw_type.return_value = {driver_name: 'fake_topic'}
            data = self.get_json('/drivers/%s/properties' % driver_name)

        self.assertEqual(mock_properties.return_value, data)
        mock_topic.assert_called_once_with(driver_name)
        mock_properties.assert_called_once_with(mock.ANY, driver_name,
                                                topic=mock_topic.return_value)
        self.assertEqual(mock_properties.return_value,
                         driver._DRIVER_PROPERTIES[driver_name])

    def test_driver_properties_cached(self, mock_topic, mock_properties):
        # only one RPC-conductor call will be made and the info cached
        # for subsequent requests
        driver._DRIVER_PROPERTIES = {}
        driver_name = 'fake'
        mock_topic.return_value = 'fake_topic'
        mock_properties.return_value = {'prop1': 'Property 1. Required.'}
        data = self.get_json('/drivers/%s/properties' % driver_name)
        data = self.get_json('/drivers/%s/properties' % driver_name)
        data = self.get_json('/drivers/%s/properties' % driver_name)
        self.assertEqual(mock_properties.return_value, data)
        mock_topic.assert_called_once_with(driver_name)
        mock_properties.assert_called_once_with(mock.ANY, driver_name,
                                                topic=mock_topic.return_value)
        self.assertEqual(mock_properties.return_value,
                         driver._DRIVER_PROPERTIES[driver_name])

    def test_driver_properties_invalid_driver_name(self, mock_topic,
                                                   mock_properties):
        # Cannot get driver properties for an invalid driver; no RPC topic
        # exists for it.
        driver._DRIVER_PROPERTIES = {}
        driver_name = 'bad_driver'
        mock_topic.side_effect = exception.DriverNotFound(
            driver_name=driver_name)
        mock_properties.return_value = {'prop1': 'Property 1. Required.'}
        ret = self.get_json('/drivers/%s/properties' % driver_name,
                            expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, ret.status_int)
        mock_topic.assert_called_once_with(driver_name)
        self.assertFalse(mock_properties.called)

    def test_driver_properties_cannot_load(self, mock_topic, mock_properties):
        # Cannot get driver properties for the driver. Although an RPC topic
        # exists for it, the conductor wasn't able to load it.
        driver._DRIVER_PROPERTIES = {}
        driver_name = 'driver'
        mock_topic.return_value = 'driver_topic'
        mock_properties.side_effect = exception.DriverNotFound(
            driver_name=driver_name)
        ret = self.get_json('/drivers/%s/properties' % driver_name,
                            expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, ret.status_int)
        mock_topic.assert_called_once_with(driver_name)
        mock_properties.assert_called_once_with(mock.ANY, driver_name,
                                                topic=mock_topic.return_value)



# Copyright (2015-2017) Hewlett Packard Enterprise Development LP
# Copyright (2015-2017) Universidade Federal de Campina Grande
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_utils import importutils

from ironic.common import driver_factory
from ironic.common import exception
from ironic.common import states
from ironic.conductor import task_manager
from ironic.drivers.modules.oneview import common
from ironic.drivers.modules.oneview import deploy_utils
from ironic import objects
from ironic.tests.unit.conductor import mgr_utils
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.db import utils as db_utils
from ironic.tests.unit.objects import utils as obj_utils

oneview_models = importutils.try_import('oneview_client.models')


@mock.patch.object(common, 'get_hponeview_client')
class OneViewDeployUtilsTestCase(db_base.DbTestCase):

    def setUp(self):
        super(OneViewDeployUtilsTestCase, self).setUp()
        self.config(manager_url='https://1.2.3.4', group='oneview')
        self.config(username='user', group='oneview')
        self.config(password='password', group='oneview')

        mgr_utils.mock_the_extension_manager(driver='fake_oneview')
        self.driver = driver_factory.get_driver('fake_oneview')

        self.node = obj_utils.create_test_node(
            self.context, driver='fake_oneview',
            properties=db_utils.get_test_oneview_properties(),
            driver_info=db_utils.get_test_oneview_driver_info(),
        )
        self.info = common.get_oneview_info(self.node)
        deploy_utils.is_node_in_use_by_oneview = mock.Mock(return_value=False)
        deploy_utils.is_node_in_use_by_ironic = mock.Mock(return_value=True)

    # Tests for prepare
    def test_prepare_node_is_in_use_by_oneview(self, mock_oneview_client):
        """`prepare` behavior when the node has a Profile on OneView."""
        deploy_utils.is_node_in_use_by_oneview.return_value = True
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.provision_state = states.DEPLOYING
            self.assertRaises(
                exception.InstanceDeployFailure,
                deploy_utils.prepare,
                task
            )

    @mock.patch.object(objects.Node, 'save')
    def test_prepare_node_is_successfuly_allocated_to_ironic(
            self, mock_save, mock_oneview_client):
        """`prepare` behavior when the node is free from OneView standpoint."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.provision_state = states.DEPLOYING
            deploy_utils.prepare(task)
            self.assertTrue(mock_save.called)

    # Tests for tear_down
    def test_tear_down(self, mock_oneview_client):
        """`tear_down` behavior when node already has Profile applied."""
        oneview_client = mock_oneview_client()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                '/rest/server-profiles/1234556789'
            )
            self.assertTrue(
                'applied_server_profile_uri' in task.node.driver_info
            )
            deploy_utils.tear_down(task)
            self.assertFalse(
                'applied_server_profile_uri' in task.node.driver_info
            )
            self.assertTrue(oneview_client.server_profiles.delete.called)

    # Tests for prepare_cleaning
    @mock.patch.object(objects.Node, 'save')
    def test_prepare_cleaning_when_node_does_not_have_sp_applied(
            self, mock_save, mock_oneview_client):
        """`prepare_cleaning` behavior when node is free."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertFalse(
                'applied_server_profile_uri' in task.node.driver_info
            )
            deploy_utils.prepare_cleaning(task)
            self.assertTrue(
                'applied_server_profile_uri' in task.node.driver_info
            )

    @mock.patch.object(objects.Node, 'save')
    def test_prepare_cleaning_when_node_has_sp_applied(
            self, mock_node_save, mock_oneview_client):
        """`prepare_cleaning` behavior when node has Profile applied."""
        oneview_client = mock_oneview_client()
        oneview_client.server_hardware.get.return_value = {
            'serverProfileUri': 'same/sp_applied'
        }
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                'same/sp_applied'
            )
            deploy_utils.prepare_cleaning(task)
            self.assertFalse(mock_node_save.called)

    def test_prepare_cleaning_node_is_in_use_by_oneview(
            self, mock_oneview_client):
        """`prepare_cleaning` behavior when node has Profile on OneView."""
        deploy_utils.is_node_in_use_by_oneview.return_value = True

        with task_manager.acquire(self.context, self.node.uuid) as task:
            self.assertRaises(
                exception.NodeCleaningFailure,
                deploy_utils.prepare_cleaning,
                task
            )

    # Tests for tear_down_cleaning
    def test_tear_down_cleaning(self, mock_oneview_client):
        """Check if Server Profile was deleted and its uri removed."""
        oneview_client = mock_oneview_client()
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                '/rest/server-profiles/1234556789'
            )
            self.assertTrue(
                'applied_server_profile_uri' in task.node.driver_info
            )
            deploy_utils.tear_down_cleaning(task)
            self.assertFalse(
                'applied_server_profile_uri' in task.node.driver_info
            )
            self.assertTrue(oneview_client.server_profiles.delete.called)

    # Tests for is_node_in_use_by_oneview
    def test_is_node_in_use_by_oneview(self, mock_oneview_client):
        """Node has a Server Profile applied by a third party user."""
        server_hardware = {
            'serverProfileUri': '/rest/server-profile/123456789'
        }
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                '/rest/server-profile/987654321'
            )
            self.assertTrue(
                deploy_utils._is_node_in_use(
                    server_hardware,
                    task.node.driver_info['applied_server_profile_uri'],
                    by_oneview=True
                )
            )

    def test_is_node_in_use_by_oneview_no_server_profile(
            self, mock_oneview_client):
        """Node has no Server Profile."""
        server_hardware = {'serverProfileUri': None}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                '/rest/server-profile/123456789'
            )
            self.assertFalse(
                deploy_utils._is_node_in_use(
                    server_hardware,
                    task.node.driver_info['applied_server_profile_uri'],
                    by_oneview=True
                )
            )

    def test_is_node_in_use_by_oneview_same_server_profile_applied(
            self, mock_oneview_client):
        """Check if node's Server Profile uri is the same applied by ironic."""
        server_hardware = {
            'serverProfileUri': '/rest/server-profile/123456789'
        }
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                '/rest/server-profile/123456789'
            )
            self.assertFalse(
                deploy_utils._is_node_in_use(
                    server_hardware,
                    task.node.driver_info['applied_server_profile_uri'],
                    by_oneview=True
                )
            )

    # Tests for is_node_in_use_by_ironic
    def test_is_node_in_use_by_ironic(self, mock_oneview_client):
        """Node has a Server Profile applied by ironic."""
        server_hardware = {'serverProfileUri': 'same/applied_sp_uri/'}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                'same/applied_sp_uri/'
            )
            self.assertTrue(
                deploy_utils._is_node_in_use(
                    server_hardware,
                    task.node.driver_info['applied_server_profile_uri'],
                    by_oneview=False
                )
            )

    def test_is_node_in_use_by_ironic_no_server_profile(
            self, mock_oneview_client):
        """Node has no Server Profile."""
        server_hardware = {'serverProfileUri': None}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                '/applied_sp_uri/'
            )
            self.assertFalse(
                deploy_utils._is_node_in_use(
                    server_hardware,
                    task.node.driver_info['applied_server_profile_uri'],
                    by_oneview=False
                )
            )

    def test__create_profile_from_template(self, mock_oneview_client):
        """Check if the server_profile was created from template."""
        server_hardware_uri = "server_hardware/12456789"
        sp_template_uri = "server_profile_template_uri/13245798"
        oneview_client = mock_oneview_client()
        oneview_client.server_profile_templates.\
            get_new_profile.return_value = {}
        server_profile = {"name": "server_profile_name",
                          "serverHardwareUri": server_hardware_uri,
                          "serverProfileTemplateUri": ""}
        deploy_utils._create_profile_from_template(
            oneview_client,
            "server_profile_name",
            server_hardware_uri,
            sp_template_uri
        )
        oneview_client.server_profiles.create.assert_called_with(
            server_profile)

    # Tests for _add_applied_server_profile_uri_field
    @mock.patch.object(objects.Node, 'save')
    def test__add_applied_server_profile_uri_field(
            self, save, mock_oneview_client):
        """Check if applied_server_profile_uri was added to driver_info."""
        server_profile = {'uri': 'any/applied_sp_uri/'}
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info.pop('applied_server_profile_uri', None)
            self.assertNotIn(
                'applied_server_profile_uri', task.node.driver_info
            )
            deploy_utils._add_applied_server_profile_uri_field(
                task.node,
                server_profile
            )
            self.assertIn('applied_server_profile_uri', task.node.driver_info)

    # Tests for _del_applied_server_profile_uri_field
    @mock.patch.object(objects.Node, 'save')
    def test__del_applied_server_profile_uri_field(
            self, save, mock_oneview_client):
        """Check if applied_server_profile_uri was removed from driver_info."""
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                'any/applied_sp_uri/'
            )
            self.assertIn('applied_server_profile_uri', task.node.driver_info)
            deploy_utils._del_applied_server_profile_uri_field(task.node)
            self.assertNotIn(
                'applied_server_profile_uri', task.node.driver_info
            )

    # Tests for allocate_server_hardware_to_ironic
    @mock.patch.object(objects.Node, 'save')
    def test_allocate_server_hardware_to_ironic(
            self, mock_node_save, mock_oneview_client):
        """Check if a Profile was created and its uri is in driver_info."""
        oneview_client = mock_oneview_client()
        server_hardware = {'serverProfileUri': None}
        oneview_client.server_hardware.get.return_value = server_hardware

        with task_manager.acquire(self.context, self.node.uuid) as task:
            deploy_utils.allocate_server_hardware_to_ironic(
                task.node, 'serverProfileName'
            )
            self.assertTrue(mock_node_save.called)
            self.assertIn('applied_server_profile_uri', task.node.driver_info)

    @mock.patch.object(objects.Node, 'save')
    def test_allocate_server_hardware_to_ironic_node_has_server_profile(
            self, mock_node_save, mock_oneview_client):
        """Test profile allocation when applied_server_profile_uri exists.

        This test consider that no Server Profile is applied on the Server
        Hardware but the applied_server_profile_uri remained on the node. Thus,
        the conductor should remove the value and apply a new server profile to
        use the node.
        """
        oneview_client = mock_oneview_client()
        server_hardware = {'serverProfileUri': None}
        oneview_client.server_hardware.get.return_value = server_hardware

        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                'any/applied_sp_uri/'
            )
            deploy_utils.allocate_server_hardware_to_ironic(
                task.node, 'serverProfileName'
            )
            self.assertTrue(mock_node_save.called)

    # Tests for deallocate_server_hardware_from_ironic
    @mock.patch.object(objects.Node, 'save')
    def test_deallocate_server_hardware_from_ironic(
            self, mock_node_save, mock_oneview_client):
        oneview_client = mock_oneview_client()
        server_hardware = {'serverProfileUri': 'any/applied_sp_uri/'}
        oneview_client.server_hardware.get.return_value = server_hardware
        with task_manager.acquire(self.context, self.node.uuid) as task:
            task.node.driver_info['applied_server_profile_uri'] = (
                'any/applied_sp_uri/'
            )
            deploy_utils.deallocate_server_hardware_from_ironic(task)
            self.assertTrue(mock_node_save.called)
            self.assertTrue(
                'applied_server_profile_uri' not in task.node.driver_info
            )



#    Copyright 2015, Cisco Systems.

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Ironic Cisco UCSM helper functions
"""

from oslo_log import log as logging
from oslo_utils import importutils
import six

from ironic.common import exception
from ironic.common.i18n import _
from ironic.drivers.modules import deploy_utils

ucs_helper = importutils.try_import('UcsSdk.utils.helper')
ucs_error = importutils.try_import('UcsSdk.utils.exception')

LOG = logging.getLogger(__name__)

REQUIRED_PROPERTIES = {
    'ucs_address': _('IP or Hostname of the UCS Manager. Required.'),
    'ucs_username': _('UCS Manager admin/server-profile username. Required.'),
    'ucs_password': _('UCS Manager password. Required.'),
    'ucs_service_profile': _('UCS Manager service-profile name. Required.')
}

COMMON_PROPERTIES = REQUIRED_PROPERTIES


def requires_ucs_client(func):
    """Creates handle to connect to UCS Manager.

    This method is being used as a decorator method. It establishes connection
    with UCS Manager. And creates a session. Any method that has to perform
    operation on UCS Manager, requries this session, which can use this method
    as decorator method. Use this method as decorator method requires having
    helper keyword argument in the definition.

    :param func: function using this as a decorator.
    :returns: a wrapper function that performs the required tasks
        mentioned above before and after calling the actual function.
    """

    @six.wraps(func)
    def wrapper(self, task, *args, **kwargs):
        if kwargs.get('helper') is None:
            kwargs['helper'] = CiscoUcsHelper(task)
        try:
            kwargs['helper'].connect_ucsm()
            return func(self, task, *args, **kwargs)
        finally:
            kwargs['helper'].logout()
    return wrapper


def parse_driver_info(node):
    """Parses and creates Cisco driver info

    :param node: An Ironic node object.
    :returns: dictonary that contains node.driver_info parameter/values.
    :raises: MissingParameterValue if any required parameters are missing.
    """

    info = {}
    for param in REQUIRED_PROPERTIES:
        info[param] = node.driver_info.get(param)
    error_msg = (_("%s driver requires these parameters to be set in the "
                   "node's driver_info.") %
                 node.driver)
    deploy_utils.check_for_missing_params(info, error_msg)
    return info


class CiscoUcsHelper(object):
    """Cisco UCS helper. Performs session managemnt."""

    def __init__(self, task):
        """Initialize with UCS Manager details.

        :param task: instance of `ironic.manager.task_manager.TaskManager`.
        """

        info = parse_driver_info(task.node)
        self.address = info['ucs_address']
        self.username = info['ucs_username']
        self.password = info['ucs_password']
        # service_profile is used by the utilities functions in UcsSdk.utils.*.
        self.service_profile = info['ucs_service_profile']
        self.handle = None
        self.uuid = task.node.uuid

    def connect_ucsm(self):
        """Creates the UcsHandle

        :raises: UcsConnectionError, if ucs helper fails to establish session
            with UCS Manager.
        """

        try:
            success, self.handle = ucs_helper.generate_ucsm_handle(
                self.address,
                self.username,
                self.password)
        except ucs_error.UcsConnectionError as ucs_exception:
            LOG.error("Cisco client: service unavailable for node "
                      "%(uuid)s.", {'uuid': self.uuid})
            raise exception.UcsConnectionError(error=ucs_exception,
                                               node=self.uuid)

    def logout(self):
        """Logouts the current active session."""

        if self.handle:
            self.handle.Logout()



# Copyright 2016 Intel Corporation
# Copyright (c) 2012 NTT DOCOMO, INC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from ironic.common.i18n import _


opts = [
    cfg.StrOpt('http_url',
               help=_("ironic-conductor node's HTTP server URL. "
                      "Example: http://192.1.2.3:8080")),
    cfg.StrOpt('http_root',
               default='/httpboot',
               help=_("ironic-conductor node's HTTP root path.")),
    cfg.IntOpt('erase_devices_priority',
               help=_('Priority to run in-band erase devices via the Ironic '
                      'Python Agent ramdisk. If unset, will use the priority '
                      'set in the ramdisk (defaults to 10 for the '
                      'GenericHardwareManager). If set to 0, will not run '
                      'during cleaning.')),
    cfg.IntOpt('erase_devices_metadata_priority',
               help=_('Priority to run in-band clean step that erases '
                      'metadata from devices, via the Ironic Python Agent '
                      'ramdisk. If unset, will use the priority set in the '
                      'ramdisk (defaults to 99 for the '
                      'GenericHardwareManager). If set to 0, will not run '
                      'during cleaning.')),
    cfg.IntOpt('shred_random_overwrite_iterations',
               default=1,
               min=0,
               help=_('During shred, overwrite all block devices N times with '
                      'random data. This is only used if a device could not '
                      'be ATA Secure Erased. Defaults to 1.')),
    cfg.BoolOpt('shred_final_overwrite_with_zeros',
                default=True,
                help=_("Whether to write zeros to a node's block devices "
                       "after writing random data. This will write zeros to "
                       "the device even when "
                       "deploy.shred_random_overwrite_iterations is 0. This "
                       "option is only used if a device could not be ATA "
                       "Secure Erased. Defaults to True.")),
    cfg.BoolOpt('continue_if_disk_secure_erase_fails',
                default=False,
                help=_('Defines what to do if an ATA secure erase operation '
                       'fails during cleaning in the Ironic Python Agent. '
                       'If False, the cleaning operation will fail and the '
                       'node will be put in ``clean failed`` state. '
                       'If True, shred will be invoked and cleaning will '
                       'continue.')),
    cfg.BoolOpt('power_off_after_deploy_failure',
                default=True,
                help=_('Whether to power off a node after deploy failure. '
                       'Defaults to True.')),
    cfg.StrOpt('default_boot_option',
               choices=['netboot', 'local'],
               help=_('Default boot option to use when no boot option is '
                      'requested in node\'s driver_info. Currently the '
                      'default is "netboot", but it will be changed to '
                      '"local" in the future. It is recommended to set '
                      'an explicit value for this option.')),
    cfg.BoolOpt('configdrive_use_object_store',
                default=False,
                deprecated_group='conductor',
                deprecated_name='configdrive_use_swift',
                help=_('Whether to upload the config drive to object store. '
                       'Set this option to True to store config drive '
                       'in swift or radosgw.')),
    cfg.StrOpt('object_store_endpoint_type',
               default='swift',
               deprecated_group='glance',
               deprecated_name='temp_url_endpoint_type',
               choices=['swift', 'radosgw'],
               help=_('Type of object store endpoint type to be '
                      'used as a backend')),
]


def register_opts(conf):
    conf.register_opts(opts, group='deploy')



#    Copyright 2017 Lenovo, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from ironic_lib import metrics_utils
from oslo_log import log as logging
from oslo_utils import importutils

from ironic.common import boot_devices
from ironic.common import exception
from ironic.common.i18n import _
from ironic.conductor import task_manager
from ironic.drivers import base
from ironic.drivers.modules.xclarity import common

LOG = logging.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)

xclarity_client_exceptions = importutils.try_import(
    'xclarity_client.exceptions')

BOOT_DEVICE_MAPPING_TO_XCLARITY = {
    boot_devices.PXE: 'PXE Network',
    boot_devices.DISK: 'Hard Disk 0',
    boot_devices.CDROM: 'CD/DVD Rom',
    boot_devices.BIOS: 'Boot To F1'
}

SUPPORTED_BOOT_DEVICES = [
    boot_devices.PXE,
    boot_devices.DISK,
    boot_devices.CDROM,
    boot_devices.BIOS,
]


class XClarityManagement(base.ManagementInterface):
    def __init__(self):
        super(XClarityManagement, self).__init__()
        self.xclarity_client = common.get_xclarity_client()

    # NOTE(TheJulia): Marking as unsupported as 3rd Party CI for this
    # interface was not established before Rocky cycle feature freeze.
    # Lenovo is continuing to work on establishing their Third Party CI,
    # and upon establishment and verification of Thid Party CI, this
    # unsupported flag shall be removed.
    # TODO(TheJulia): If Third Party CI is not online prior to the
    # Stein Feature Freeze, this interface should be removed.
    supported = False

    def get_properties(self):
        return common.COMMON_PROPERTIES

    @METRICS.timer('XClarityManagement.validate')
    def validate(self, task):
        """It validates if the node is being used by XClarity.

        :param task: a task from TaskManager.
        """
        common.is_node_managed_by_xclarity(self.xclarity_client, task.node)

    @METRICS.timer('XClarityManagement.get_supported_boot_devices')
    def get_supported_boot_devices(self, task):
        """Gets a list of the supported boot devices.

        :param task: a task from TaskManager.
        :returns: A list with the supported boot devices defined
                  in :mod:`ironic.common.boot_devices`.
        """

        return SUPPORTED_BOOT_DEVICES

    def _validate_supported_boot_device(self, task, boot_device):
        """It validates if the boot device is supported by XClarity.

        :param task: a task from TaskManager.
        :param boot_device: the boot device, one of [PXE, DISK, CDROM, BIOS]
        :raises: InvalidParameterValue if the boot device is not supported.
        """
        if boot_device not in SUPPORTED_BOOT_DEVICES:
            raise exception.InvalidParameterValue(
                _("Unsupported boot device %(device)s for node: %(node)s ")
                % {"device": boot_device, "node": task.node.uuid}
            )

    @METRICS.timer('XClarityManagement.get_boot_device')
    def get_boot_device(self, task):
        """Get the current boot device for the task's node.

        :param task: a task from TaskManager.
        :returns: a dictionary containing:
            :boot_device: the boot device, one of [PXE, DISK, CDROM, BIOS]
            :persistent: Whether the boot device will persist or not
        :raises: InvalidParameterValue if the boot device is unknown
        :raises: XClarityError if the communication with XClarity fails
        """
        server_hardware_id = common.get_server_hardware_id(task.node)
        try:
            boot_info = (
                self.xclarity_client.get_node_all_boot_info(
                    server_hardware_id)
            )
        except xclarity_client_exceptions.XClarityError as xclarity_exc:
            LOG.error(
                "Error getting boot device from XClarity for node %(node)s. "
                "Error: %(error)s", {'node': task.node.uuid,
                                     'error': xclarity_exc})
            raise common.XClarityError(error=xclarity_exc)

        persistent = False
        primary = None
        boot_order = boot_info['bootOrder']['bootOrderList']
        for item in boot_order:
            current = item.get('currentBootOrderDevices', None)
            boot_type = item.get('bootType', None)
            if boot_type == "SingleUse":
                persistent = False
                primary = current[0]
                if primary != 'None':
                    boot_device = {'boot_device': primary,
                                   'persistent': persistent}
                    self._validate_whether_supported_boot_device(primary)
                    return boot_device
            elif boot_type == "Permanent":
                persistent = True
                boot_device = {'boot_device': current[0],
                               'persistent': persistent}
                self._validate_supported_boot_device(task, primary)
                return boot_device

    @METRICS.timer('XClarityManagement.set_boot_device')
    @task_manager.require_exclusive_lock
    def set_boot_device(self, task, device, persistent=False):
        """Sets the boot device for a node.

        :param task: a task from TaskManager.
        :param device: the boot device, one of the supported devices
                       listed in :mod:`ironic.common.boot_devices`.
        :param persistent: Boolean value. True if the boot device will
                           persist to all future boots, False if not.
                           Default: False.
        :raises: InvalidParameterValue if an invalid boot device is
                 specified.
        :raises: XClarityError if the communication with XClarity fails
        """
        self._validate_supported_boot_device(task=task, boot_device=device)

        server_hardware_id = task.node.driver_info.get('server_hardware_id')
        LOG.debug("Setting boot device to %(device)s for node %(node)s",
                  {"device": device, "node": task.node.uuid})
        self._set_boot_device(task, server_hardware_id, device,
                              singleuse=not persistent)

    @METRICS.timer('XClarityManagement.get_sensors_data')
    def get_sensors_data(self, task):
        """Get sensors data.

        :param task: a TaskManager instance.
        :raises: NotImplementedError

        """
        raise NotImplementedError()

    def _translate_ironic_to_xclarity(self, boot_device):
        """Translates Ironic boot options to Xclarity boot options.

        :param boot_device: Ironic boot_device
        :returns: Translated XClarity boot_device.

        """
        return BOOT_DEVICE_MAPPING_TO_XCLARITY.get(boot_device)

    def _set_boot_device(self, task, server_hardware_id,
                         new_primary_boot_device, singleuse=False):
        """Set the current boot device for xclarity

        :param server_hardware_id: the uri of the server hardware in XClarity
        :param new_primary_boot_device: boot device to be set
        :param task: a TaskManager instance.
        :param singleuse: if this device will be used only once at next boot
        """
        boot_info = self.xclarity_client.get_node_all_boot_info(
            server_hardware_id)
        xclarity_boot_device = self._translate_ironic_to_xclarity(
            new_primary_boot_device)
        current = []
        LOG.debug(
            ("Setting boot device to %(device)s for XClarity "
             "node %(node)s"),
            {'device': xclarity_boot_device, 'node': task.node.uuid}
        )
        for item in boot_info['bootOrder']['bootOrderList']:
            if singleuse and item['bootType'] == 'SingleUse':
                item['currentBootOrderDevices'][0] = xclarity_boot_device
            elif not singleuse and item['bootType'] == 'Permanent':
                current = item['currentBootOrderDevices']
                if xclarity_boot_device == current[0]:
                    return
                if xclarity_boot_device in current:
                    current.remove(xclarity_boot_device)
                current.insert(0, xclarity_boot_device)
                item['currentBootOrderDevices'] = current

        try:
            self.xclarity_client.set_node_boot_info(server_hardware_id,
                                                    boot_info,
                                                    xclarity_boot_device,
                                                    singleuse)
        except xclarity_client_exceptions.XClarityError as xclarity_exc:
            LOG.error(
                ('Error setting boot device %(boot_device)s for the XClarity '
                 'node %(node)s. Error: %(error)s'),
                {'boot_device': xclarity_boot_device, 'node': task.node.uuid,
                 'error': xclarity_exc}
            )
            raise common.XClarityError(error=xclarity_exc)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DRAC Lifecycle job specific methods
"""

from oslo_log import log as logging
from oslo_utils import importutils

from ironic.common import exception
from ironic.common.i18n import _
from ironic.drivers.modules.drac import common as drac_common

drac_exceptions = importutils.try_import('dracclient.exceptions')

LOG = logging.getLogger(__name__)


def validate_job_queue(node):
    """Validates the job queue on the node.

    It raises an exception if an unfinished configuration job exists.

    :param node: an ironic node object.
    :raises: DracOperationError on an error from python-dracclient.
    """

    unfinished_jobs = list_unfinished_jobs(node)
    if unfinished_jobs:
        msg = _('Unfinished config jobs found: %(jobs)r. Make sure they are '
                'completed before retrying.') % {'jobs': unfinished_jobs}
        raise exception.DracOperationError(error=msg)


def get_job(node, job_id):
    """Get the details of a Lifecycle job of the node.

    :param node: an ironic node object.
    :param job_id: ID of the Lifecycle job.
    :returns: a Job object from dracclient.
    :raises: DracOperationError on an error from python-dracclient.
    """
    client = drac_common.get_drac_client(node)

    try:
        return client.get_job(job_id)
    except drac_exceptions.BaseClientException as exc:
        LOG.error('DRAC driver failed to get the job %(job_id)s '
                  'for node %(node_uuid)s. Reason: %(error)s.',
                  {'node_uuid': node.uuid,
                   'error': exc})
        raise exception.DracOperationError(error=exc)


def list_unfinished_jobs(node):
    """List unfinished config jobs of the node.

    :param node: an ironic node object.
    :returns: a list of Job objects from dracclient.
    :raises: DracOperationError on an error from python-dracclient.
    """
    client = drac_common.get_drac_client(node)

    try:
        return client.list_jobs(only_unfinished=True)
    except drac_exceptions.BaseClientException as exc:
        LOG.error('DRAC driver failed to get the list of unfinished jobs '
                  'for node %(node_uuid)s. Reason: %(error)s.',
                  {'node_uuid': node.uuid,
                   'error': exc})
        raise exception.DracOperationError(error=exc)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Flat network interface. Useful for shared, flat networks.
"""

from neutronclient.common import exceptions as neutron_exceptions
from oslo_config import cfg
from oslo_log import log

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import neutron
from ironic.drivers import base
from ironic.drivers.modules.network import common


LOG = log.getLogger(__name__)

CONF = cfg.CONF


class FlatNetwork(common.NeutronVIFPortIDMixin,
                  neutron.NeutronNetworkInterfaceMixin, base.NetworkInterface):
    """Flat network interface."""

    def __init__(self):
        cleaning_net = CONF.neutron.cleaning_network
        if not cleaning_net:
            LOG.warning(
                'Please specify a valid UUID or name for '
                '[neutron]/cleaning_network configuration option so that '
                'this interface is able to perform cleaning. Otherwise, '
                'cleaning operations will fail to start.')

    def validate(self, task):
        """Validates the network interface.

        :param task: a TaskManager instance.
        :raises: InvalidParameterValue, if the network interface configuration
            is invalid.
        :raises: MissingParameterValue, if some parameters are missing.
        """
        self.get_cleaning_network_uuid(task)

    def _bind_flat_ports(self, task):
        LOG.debug("Binding flat network ports")
        client = neutron.get_client(context=task.context)
        for port_like_obj in task.ports + task.portgroups:
            vif_port_id = (
                port_like_obj.internal_info.get(common.TENANT_VIF_KEY) or
                port_like_obj.extra.get('vif_port_id')
            )
            if not vif_port_id:
                continue
            body = {
                'port': {
                    'binding:host_id': task.node.uuid,
                    'binding:vnic_type': neutron.VNIC_BAREMETAL,
                    'mac_address': port_like_obj.address
                }
            }
            try:
                client.update_port(vif_port_id, body)
            except neutron_exceptions.NeutronClientException as e:
                msg = (_('Unable to set binding:host_id for '
                         'neutron port %(port_id)s. Error: '
                         '%(err)s') % {'port_id': vif_port_id, 'err': e})
                LOG.exception(msg)
                raise exception.NetworkError(msg)

    def _unbind_flat_ports(self, task):
        node = task.node
        LOG.info('Unbinding instance ports from node %s', node.uuid)

        ports = [p for p in task.ports if not p.portgroup_id]
        portgroups = task.portgroups
        for port_like_obj in ports + portgroups:
            vif_port_id = (
                port_like_obj.internal_info.get(common.TENANT_VIF_KEY) or
                port_like_obj.extra.get('vif_port_id'))
            if not vif_port_id:
                continue
            neutron.unbind_neutron_port(vif_port_id, context=task.context)

    def add_provisioning_network(self, task):
        """Add the provisioning network to a node.

        :param task: A TaskManager instance.
        :raises: NetworkError when failed to set binding:host_id
        """
        self._bind_flat_ports(task)

    def remove_provisioning_network(self, task):
        """Remove the provisioning network from a node.

        :param task: A TaskManager instance.
        """
        self._unbind_flat_ports(task)

    def configure_tenant_networks(self, task):
        """Configure tenant networks for a node.

        :param task: A TaskManager instance.
        """
        self._bind_flat_ports(task)

    def unconfigure_tenant_networks(self, task):
        """Unconfigure tenant networks for a node.

        Unbind the port here/now to avoid the possibility of the ironic port
        being bound to the tenant and cleaning networks at the same time.

        :param task: A TaskManager instance.
        :raises: NetworkError
        """
        self._unbind_flat_ports(task)

    def add_cleaning_network(self, task):
        """Add the cleaning network to a node.

        :param task: A TaskManager instance.
        :returns: a dictionary in the form {port.uuid: neutron_port['id']}
        :raises: NetworkError, InvalidParameterValue
        """
        # If we have left over ports from a previous cleaning, remove them
        neutron.rollback_ports(task,
                               self.get_cleaning_network_uuid(task))
        LOG.info('Adding cleaning network to node %s', task.node.uuid)
        vifs = neutron.add_ports_to_network(
            task, self.get_cleaning_network_uuid(task))
        for port in task.ports:
            if port.uuid in vifs:
                internal_info = port.internal_info
                internal_info['cleaning_vif_port_id'] = vifs[port.uuid]
                port.internal_info = internal_info
                port.save()
        return vifs

    def remove_cleaning_network(self, task):
        """Remove the cleaning network from a node.

        :param task: A TaskManager instance.
        :raises: NetworkError
        """
        LOG.info('Removing ports from cleaning network for node %s',
                 task.node.uuid)
        neutron.remove_ports_from_network(
            task, self.get_cleaning_network_uuid(task))
        for port in task.ports:
            if 'cleaning_vif_port_id' in port.internal_info:
                internal_info = port.internal_info
                del internal_info['cleaning_vif_port_id']
                port.internal_info = internal_info
                port.save()

    def add_rescuing_network(self, task):
        """Add the rescuing network to a node.

        Flat network does not use the rescuing network.
        Bind the port again since unconfigure_tenant_network() unbound it.

        :param task: A TaskManager instance.
        :returns: a dictionary in the form {port.uuid: neutron_port['id']}
        :raises: NetworkError, InvalidParameterValue
        """
        LOG.info('Bind ports for rescuing node %s', task.node.uuid)
        self._bind_flat_ports(task)

    def remove_rescuing_network(self, task):
        """Remove the rescuing network from a node.

        Flat network does not use the rescuing network.
        Unbind the port again since add_rescuing_network() bound it.

        :param task: A TaskManager instance.
        :raises: NetworkError
        """
        LOG.info('Unbind ports for rescuing node %s', task.node.uuid)
        self._unbind_flat_ports(task)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Set Port.pxe_enabled to True if NULL

Revision ID: f6fdb920c182
Revises: 5ea1b0d310e
Create Date: 2016-02-12 16:53:21.008580

"""

# revision identifiers, used by Alembic.
revision = 'f6fdb920c182'
down_revision = '5ea1b0d310e'

from alembic import op
from sqlalchemy import Boolean, String
from sqlalchemy.sql import table, column, null

port = table('ports',
             column('uuid', String(36)),
             column('pxe_enabled', Boolean()))


def upgrade():
    op.execute(
        port.update().where(
            port.c.pxe_enabled == null()).values(
                {'pxe_enabled': True}))



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic_lib import metrics_utils
import pecan
from pecan import rest
from six.moves import http_client
import wsme
from wsme import types as wtypes

from ironic.api.controllers import base
from ironic.api.controllers import link
from ironic.api.controllers.v1 import types
from ironic.api.controllers.v1 import utils as api_utils
from ironic.api import expose
from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import policy
from ironic.drivers import base as driver_base


METRICS = metrics_utils.get_metrics_logger(__name__)

# Property information for drivers:
#   key = driver name;
#   value = dictionary of properties of that driver:
#             key = property name.
#             value = description of the property.
# NOTE(rloo). This is cached for the lifetime of the API service. If one or
# more conductor services are restarted with new driver versions, the API
# service should be restarted.
_DRIVER_PROPERTIES = {}

# Vendor information for drivers:
#   key = driver name;
#   value = dictionary of vendor methods of that driver:
#             key = method name.
#             value = dictionary with the metadata of that method.
# NOTE(lucasagomes). This is cached for the lifetime of the API
# service. If one or more conductor services are restarted with new driver
# versions, the API service should be restarted.
_VENDOR_METHODS = {}

# RAID (logical disk) configuration information for drivers:
#   key = driver name;
#   value = dictionary of RAID configuration information of that driver:
#             key = property name.
#             value = description of the property
# NOTE(rloo). This is cached for the lifetime of the API service. If one or
# more conductor services are restarted with new driver versions, the API
# service should be restarted.
_RAID_PROPERTIES = {}


def hide_fields_in_newer_versions(obj):
    """This method hides fields that were added in newer API versions.

    Certain fields were introduced at certain API versions.
    These fields are only made available when the request's API version
    matches or exceeds the versions when these fields were introduced.
    """
    if not api_utils.allow_storage_interface():
        obj.default_storage_interface = wsme.Unset
        obj.enabled_storage_interfaces = wsme.Unset

    if not api_utils.allow_rescue_interface():
        obj.default_rescue_interface = wsme.Unset
        obj.enabled_rescue_interfaces = wsme.Unset


class Driver(base.APIBase):
    """API representation of a driver."""

    name = wtypes.text
    """The name of the driver"""

    hosts = [wtypes.text]
    """A list of active conductors that support this driver"""

    type = wtypes.text
    """Whether the driver is classic or dynamic (hardware type)"""

    links = wsme.wsattr([link.Link], readonly=True)
    """A list containing self and bookmark links"""

    properties = wsme.wsattr([link.Link], readonly=True)
    """A list containing links to driver properties"""

    """Default interface for a hardware type"""
    default_boot_interface = wtypes.text
    default_console_interface = wtypes.text
    default_deploy_interface = wtypes.text
    default_inspect_interface = wtypes.text
    default_management_interface = wtypes.text
    default_network_interface = wtypes.text
    default_power_interface = wtypes.text
    default_raid_interface = wtypes.text
    default_rescue_interface = wtypes.text
    default_storage_interface = wtypes.text
    default_vendor_interface = wtypes.text

    """A list of enabled interfaces for a hardware type"""
    enabled_boot_interfaces = [wtypes.text]
    enabled_console_interfaces = [wtypes.text]
    enabled_deploy_interfaces = [wtypes.text]
    enabled_inspect_interfaces = [wtypes.text]
    enabled_management_interfaces = [wtypes.text]
    enabled_network_interfaces = [wtypes.text]
    enabled_power_interfaces = [wtypes.text]
    enabled_raid_interfaces = [wtypes.text]
    enabled_rescue_interfaces = [wtypes.text]
    enabled_storage_interfaces = [wtypes.text]
    enabled_vendor_interfaces = [wtypes.text]

    @staticmethod
    def convert_with_links(name, hosts, driver_type, detail=False,
                           interface_info=None):
        """Convert driver/hardware type info to an API-serializable object.

        :param name: name of driver or hardware type.
        :param hosts: list of conductor hostnames driver is active on.
        :param driver_type: 'classic' for classic drivers, 'dynamic' for
                            hardware types.
        :param detail: boolean, whether to include detailed info, such as
                       the 'type' field and default/enabled interfaces fields.
        :param interface_info: optional list of dicts of hardware interface
                               info.
        :returns: API-serializable driver object.
        """
        driver = Driver()
        driver.name = name
        driver.hosts = hosts
        driver.links = [
            link.Link.make_link('self',
                                pecan.request.public_url,
                                'drivers', name),
            link.Link.make_link('bookmark',
                                pecan.request.public_url,
                                'drivers', name,
                                bookmark=True)
        ]
        if api_utils.allow_links_node_states_and_driver_properties():
            driver.properties = [
                link.Link.make_link('self',
                                    pecan.request.public_url,
                                    'drivers', name + "/properties"),
                link.Link.make_link('bookmark',
                                    pecan.request.public_url,
                                    'drivers', name + "/properties",
                                    bookmark=True)
            ]

        if api_utils.allow_dynamic_drivers():
            driver.type = driver_type
            if driver_type == 'dynamic' and detail:
                if interface_info is None:
                    # TODO(jroll) objectify this
                    interface_info = (pecan.request.dbapi
                                      .list_hardware_type_interfaces([name]))
                for iface_type in driver_base.ALL_INTERFACES:
                    default = None
                    enabled = set()
                    for iface in interface_info:
                        if iface['interface_type'] == iface_type:
                            iface_name = iface['interface_name']
                            enabled.add(iface_name)
                            # NOTE(jroll) this assumes the default is the same
                            # on all conductors
                            if iface['default']:
                                default = iface_name

                    default_key = 'default_%s_interface' % iface_type
                    enabled_key = 'enabled_%s_interfaces' % iface_type
                    setattr(driver, default_key, default)
                    setattr(driver, enabled_key, list(enabled))

            elif detail:
                for iface_type in driver_base.ALL_INTERFACES:
                    # always return None for classic drivers
                    setattr(driver, 'default_%s_interface' % iface_type, None)
                    setattr(driver, 'enabled_%s_interfaces' % iface_type, None)

        hide_fields_in_newer_versions(driver)
        return driver

    @classmethod
    def sample(cls):
        attrs = {
            'name': 'sample-driver',
            'hosts': ['fake-host'],
            'type': 'classic',
        }
        for iface_type in driver_base.ALL_INTERFACES:
            attrs['default_%s_interface' % iface_type] = None
            attrs['enabled_%s_interfaces' % iface_type] = None

        sample = cls(**attrs)
        return sample


class DriverList(base.APIBase):
    """API representation of a list of drivers."""

    drivers = [Driver]
    """A list containing drivers objects"""

    @staticmethod
    def convert_with_links(drivers, hardware_types, detail=False):
        """Convert drivers and hardware types to an API-serializable object.

        :param drivers: dict mapping driver names to conductor hostnames.
        :param hardware_types: dict mapping hardware type names to conductor
                               hostnames.
        :param detail: boolean, whether to include detailed info, such as
                       the 'type' field and default/enabled interfaces fields.
        :returns: an API-serializable driver collection object.
        """
        collection = DriverList()
        collection.drivers = [
            Driver.convert_with_links(dname, list(drivers[dname]), 'classic',
                                      detail=detail)
            for dname in drivers]

        # NOTE(jroll) we return hardware types in all API versions,
        # but restrict type/default/enabled fields to 1.30.
        # This is checked in Driver.convert_with_links(), however also
        # checking here can save us a DB query.
        if api_utils.allow_dynamic_drivers() and detail:
            iface_info = pecan.request.dbapi.list_hardware_type_interfaces(
                list(hardware_types))
        else:
            iface_info = []

        for htname in hardware_types:
            interface_info = [i for i in iface_info
                              if i['hardware_type'] == htname]
            collection.drivers.append(
                Driver.convert_with_links(htname,
                                          list(hardware_types[htname]),
                                          'dynamic', detail=detail,
                                          interface_info=interface_info))
        return collection

    @classmethod
    def sample(cls):
        sample = cls()
        sample.drivers = [Driver.sample()]
        return sample


class DriverPassthruController(rest.RestController):
    """REST controller for driver passthru.

    This controller allow vendors to expose cross-node functionality in the
    Ironic API. Ironic will merely relay the message from here to the specified
    driver, no introspection will be made in the message body.
    """

    _custom_actions = {
        'methods': ['GET']
    }

    @METRICS.timer('DriverPassthruController.methods')
    @expose.expose(wtypes.text, wtypes.text)
    def methods(self, driver_name):
        """Retrieve information about vendor methods of the given driver.

        :param driver_name: name of the driver.
        :returns: dictionary with <vendor method name>:<method metadata>
                  entries.
        :raises: DriverNotFound if the driver name is invalid or the
                 driver cannot be loaded.
        """
        cdict = pecan.request.context.to_policy_values()
        policy.authorize('baremetal:driver:vendor_passthru', cdict, cdict)

        if driver_name not in _VENDOR_METHODS:
            topic = pecan.request.rpcapi.get_topic_for_driver(driver_name)
            ret = pecan.request.rpcapi.get_driver_vendor_passthru_methods(
                pecan.request.context, driver_name, topic=topic)
            _VENDOR_METHODS[driver_name] = ret

        return _VENDOR_METHODS[driver_name]

    @METRICS.timer('DriverPassthruController._default')
    @expose.expose(wtypes.text, wtypes.text, wtypes.text,
                   body=wtypes.text)
    def _default(self, driver_name, method, data=None):
        """Call a driver API extension.

        :param driver_name: name of the driver to call.
        :param method: name of the method, to be passed to the vendor
                       implementation.
        :param data: body of data to supply to the specified method.
        """
        cdict = pecan.request.context.to_policy_values()
        policy.authorize('baremetal:driver:vendor_passthru', cdict, cdict)

        topic = pecan.request.rpcapi.get_topic_for_driver(driver_name)
        return api_utils.vendor_passthru(driver_name, method, topic, data=data,
                                         driver_passthru=True)


class DriverRaidController(rest.RestController):

    _custom_actions = {
        'logical_disk_properties': ['GET']
    }

    @METRICS.timer('DriverRaidController.logical_disk_properties')
    @expose.expose(types.jsontype, wtypes.text)
    def logical_disk_properties(self, driver_name):
        """Returns the logical disk properties for the driver.

        :param driver_name: Name of the driver.
        :returns: A dictionary containing the properties that can be mentioned
            for logical disks and a textual description for them.
        :raises: UnsupportedDriverExtension if the driver doesn't
            support RAID configuration.
        :raises: NotAcceptable, if requested version of the API is less than
            1.12.
        :raises: DriverNotFound, if driver is not loaded on any of the
            conductors.
        """
        cdict = pecan.request.context.to_policy_values()
        policy.authorize('baremetal:driver:get_raid_logical_disk_properties',
                         cdict, cdict)

        if not api_utils.allow_raid_config():
            raise exception.NotAcceptable()

        if driver_name not in _RAID_PROPERTIES:
            topic = pecan.request.rpcapi.get_topic_for_driver(driver_name)
            try:
                info = pecan.request.rpcapi.get_raid_logical_disk_properties(
                    pecan.request.context, driver_name, topic=topic)
            except exception.UnsupportedDriverExtension as e:
                # Change error code as 404 seems appropriate because RAID is a
                # standard interface and all drivers might not have it.
                e.code = http_client.NOT_FOUND
                raise

            _RAID_PROPERTIES[driver_name] = info
        return _RAID_PROPERTIES[driver_name]


class DriversController(rest.RestController):
    """REST controller for Drivers."""

    vendor_passthru = DriverPassthruController()

    raid = DriverRaidController()
    """Expose RAID as a sub-element of drivers"""

    _custom_actions = {
        'properties': ['GET'],
    }

    @METRICS.timer('DriversController.get_all')
    @expose.expose(DriverList, wtypes.text, types.boolean)
    def get_all(self, type=None, detail=None):
        """Retrieve a list of drivers."""
        # FIXME(deva): formatting of the auto-generated REST API docs
        #              will break from a single-line doc string.
        #              This is a result of a bug in sphinxcontrib-pecanwsme
        # https://github.com/dreamhost/sphinxcontrib-pecanwsme/issues/8
        cdict = pecan.request.context.to_policy_values()
        policy.authorize('baremetal:driver:get', cdict, cdict)

        api_utils.check_allow_driver_detail(detail)
        api_utils.check_allow_filter_driver_type(type)
        if type not in (None, 'classic', 'dynamic'):
            raise exception.Invalid(_(
                '"type" filter must be one of "classic" or "dynamic", '
                'if specified.'))

        driver_list = {}
        hw_type_dict = {}
        if type is None or type == 'classic':
            driver_list = pecan.request.dbapi.get_active_driver_dict()
        if type is None or type == 'dynamic':
            hw_type_dict = pecan.request.dbapi.get_active_hardware_type_dict()
        return DriverList.convert_with_links(driver_list, hw_type_dict,
                                             detail=detail)

    @METRICS.timer('DriversController.get_one')
    @expose.expose(Driver, wtypes.text)
    def get_one(self, driver_name):
        """Retrieve a single driver."""
        # NOTE(russell_h): There is no way to make this more efficient than
        # retrieving a list of drivers using the current sqlalchemy schema, but
        # this path must be exposed for Pecan to route any paths we might
        # choose to expose below it.
        cdict = pecan.request.context.to_policy_values()
        policy.authorize('baremetal:driver:get', cdict, cdict)

        def _find_driver(driver_dict, driver_type):
            for name, hosts in driver_dict.items():
                if name == driver_name:
                    return Driver.convert_with_links(name, list(hosts),
                                                     driver_type, detail=True)

        hw_type_dict = pecan.request.dbapi.get_active_hardware_type_dict()
        driver = _find_driver(hw_type_dict, 'dynamic')
        if driver:
            return driver
        driver_dict = pecan.request.dbapi.get_active_driver_dict()
        driver = _find_driver(driver_dict, 'classic')
        if driver:
            return driver

        raise exception.DriverNotFound(driver_name=driver_name)

    @METRICS.timer('DriversController.properties')
    @expose.expose(wtypes.text, wtypes.text)
    def properties(self, driver_name):
        """Retrieve property information of the given driver.

        :param driver_name: name of the driver.
        :returns: dictionary with <property name>:<property description>
                  entries.
        :raises: DriverNotFound (HTTP 404) if the driver name is invalid or
                 the driver cannot be loaded.
        """
        cdict = pecan.request.context.to_policy_values()
        policy.authorize('baremetal:driver:get_properties', cdict, cdict)

        if driver_name not in _DRIVER_PROPERTIES:
            topic = pecan.request.rpcapi.get_topic_for_driver(driver_name)
            properties = pecan.request.rpcapi.get_driver_properties(
                pecan.request.context, driver_name, topic=topic)
            _DRIVER_PROPERTIES[driver_name] = properties

        return _DRIVER_PROPERTIES[driver_name]
